# -*- coding: utf-8 -*-
from collections import namedtuple, Iterable
from typing import Iterable, Union
from functools import partial

from scipy.sparse import coo_matrix
import numpy as np
from numpy import ndarray

from neumann import squeeze, config
from neumann.linalg import ReferenceFrame
from neumann.array import atleast1d, atleastnd, ascont
from neumann.utils import to_range_1d
from neumann.linalg.sparse.jaggedarray import JaggedArray
from polymesh.topo import TopologyArray

from ..preproc import (fem_coeff_matrix_coo, assert_min_diagonals_bulk,
                       assemble_load_vector, condensate_Kf_bulk, condensate_M_bulk)
from ..postproc import (approx_element_solution_bulk, calculate_external_forces_bulk,
                        calculate_internal_forces_bulk, explode_kinetic_strains,
                        element_dof_solution_bulk)
from ..utils import (topo_to_gnum, expand_coeff_matrix_bulk, element_dofmap_bulk,
                     topo_to_gnum_jagged)
from ..tr import (nodal_dcm, nodal_dcm_bulk, element_dcm, element_dcm_bulk,
                  tr_element_vectors_bulk_multi as tr1d, tr_element_matrices_bulk as tr2d)
from .utils import (stiffness_matrix_bulk2, strain_displacement_matrix_bulk2,
                    unit_strain_load_vector_bulk, strain_load_vector_bulk, mass_matrix_bulk)
from .metaelem import FemMixin
from .celldata import CellData


Quadrature = namedtuple('QuadratureRule', ['inds', 'pos', 'weight'])


UX, UY, UZ, ROTX, ROTY, ROTZ = FX, FY, FZ, MX, MY, MZ = list(range(6))
ulabels = {UX: 'UX', UY: 'UY', UZ: 'UZ',
           ROTX: 'ROTX', ROTY: 'ROTY', ROTZ: 'ROTZ'}
flabels = {FX: 'FX', FY: 'FY', FZ: 'FZ',
           MX: 'MX', MY: 'MY', MZ: 'MZ'}
ulabel_to_int = {v: k for k, v in ulabels.items()}
flabel_to_int = {v: k for k, v in flabels.items()}


class FiniteElement(CellData, FemMixin):
    """
    Base mixin class for all element types. Functions of this class
    can be called on any class of SigmaEpsilon.

    """

    def direction_cosine_matrix(self, *args, source: Union[ndarray, str] = None,
                                target: Union[ndarray, str] = None, N: int = None,
                                **kwargs) -> ndarray:
        """
        Returns the DCM matrix from local to global for all elements 
        in the block.

        Parameters
        ----------
        source : str or ReferenceFrame, Optional
            A source frame. Default is None.

        target : str or ReferenceFrame, Optional
            A target frame. Default is None.

        N : int, Optional
            Number of points. If not specified, the number of nodes is inferred from 
            the class of the instance the function is called upon. Default is None.

        Returns
        -------
        numpy.ndarray
            The dcm matrix for linear transformations from source to target.

        """
        nNE = self.__class__.NNODE if N is None else N
        nDOF = self.container.root().NDOFN
        c, r = divmod(nDOF, 3)
        assert r == 0, "The default mechanism assumes that the number of " + \
            "deegrees of freedom per node is a multiple of 3."
        ndcm = nodal_dcm_bulk(self.frames, c)
        dcm = element_dcm_bulk(ndcm, nNE, nDOF)  # (nE, nEVAB, nEVAB)
        if source is None and target is None:
            target = 'global'
        if source is not None:
            if isinstance(source, str):
                if source == 'global':
                    S = ReferenceFrame(dim=dcm.shape[-1])
                    return ReferenceFrame(dcm).dcm(source=S)
            elif isinstance(source, ReferenceFrame):
                if len(source) == 3:
                    c, r = divmod(nDOF, 3)
                    assert r == 0, "The default mechanism assumes that the number of " + \
                        "deegrees of freedom per node is a multiple of 3."
                    s_ndcm = nodal_dcm(source.dcm(), c)
                    s_dcm = element_dcm(s_ndcm, nNE, nDOF)  # (nEVAB, nEVAB)
                    source = ReferenceFrame(s_dcm)
                return ReferenceFrame(dcm).dcm(source=source)
            else:
                raise NotImplementedError
        elif target is not None:
            if isinstance(target, str):
                if target == 'global':
                    T = ReferenceFrame(dim=dcm.shape[-1])
                    return ReferenceFrame(dcm).dcm(target=T)
            elif isinstance(target, ReferenceFrame):
                if len(target) == 3:
                    c, r = divmod(nDOF, 3)
                    assert r == 0, "The default mechanism assumes that the number of " + \
                        "deegrees of freedom per node is a multiple of 3."
                    t_ndcm = nodal_dcm(target.dcm(), c)
                    t_dcm = element_dcm(t_ndcm, nNE, nDOF)  # (nEVAB, nEVAB)
                    target = ReferenceFrame(t_dcm)
                return ReferenceFrame(dcm).dcm(target=target)
            else:
                raise NotImplementedError
        raise NotImplementedError

    @squeeze(True)
    def dof_solution(self, *args, target: Union[str, ReferenceFrame] = 'local',
                     cells: Union[int, Iterable[int]] = None,
                     points: Union[float, Iterable] = None,
                     rng: Iterable = None, flatten: bool = True, **kwargs) -> ndarray:
        """
        Returns nodal displacements for the cells, wrt. their local frames.

        Parameters
        ----------
        points : float or Iterable, Optional
                Points of evaluation. If provided, it is assumed that the given values
                are wrt. the range [0, 1], unless specified otherwise with the 'rng' 
                parameter. If not provided, results are returned for the nodes of the 
                selected elements. Default is None.

        rng : Iterable, Optional
            Range where the points of evauation are understood. Default is [0, 1].

        cells : int or Iterable[int], Optional
            Indices of cells. If not provided, results are returned for all cells.
            If cells are provided, the function returns a dictionary, with the cell 
            indices being the keys. Default is None.

        target : str or ReferenceFrame, Optional
            Reference frame for the output. A value of None or 'local' refers to the 
            local system of the cells. Default is 'local'.

        Returns
        -------
        numpy.ndarray
            An array of shape (nE, nEVAB, nRHS) if 'flatten' is True 
            else (nE, nNE, nDOF, nRHS).

        """
        if cells is not None:
            cells = atleast1d(cells)
            conds = np.isin(cells, self.id)
            cells = atleast1d(cells[conds])
            assert len(cells) > 0, "Length of cells is zero!"
        else:
            cells = np.s_[:]

        dofsol = self.pointdata.dofsol
        dofsol = atleastnd(dofsol, 3, back=True)
        nP, nDOF, nRHS = dofsol.shape
        dofsol = dofsol.reshape(nP * nDOF, nRHS)
        gnum = self.global_dof_numbering()[cells]

        # transform values to cell-local frames
        dcm = self.direction_cosine_matrix(source='global')[cells]
        values = element_dof_solution_bulk(dofsol, gnum)  # (nE, nEVAB, nRHS)
        values = ascont(np.swapaxes(values, 1, 2))  # (nE, nRHS, nEVAB)
        values = tr1d(values, dcm)
        values = ascont(np.swapaxes(values, 1, 2))  # (nE, nEVAB, nRHS)

        if points is None:
            points = np.array(self.lcoords()).flatten()
            rng = [-1, 1]
        else:
            rng = np.array([0, 1]) if rng is None else np.array(rng)

        # approximate at points
        # values -> (nE, nEVAB, nRHS)
        points, rng = to_range_1d(
            points, source=rng, target=[-1, 1]).flatten(), [-1, 1]

        N = self.shape_function_matrix(points, rng=rng)
        # N -> (nP, nDOF, nDOF * nNODE)
        values = ascont(np.swapaxes(values, 1, 2))  # (nE, nRHS, nEVAB)
        values = approx_element_solution_bulk(values, N)
        # values -> (nE, nRHS, nP, nDOF)

        if target is not None:
            # transform values to a destination frame, otherwise return
            # the results in the local frames of the cells
            if isinstance(target, str) and target == 'local':
                pass
            else:
                nE, nRHS, nP, nDOF = values.shape
                values = values.reshape(nE, nRHS, nP * nDOF)
                dcm = self.direction_cosine_matrix(N=nP, target=target)[cells]
                values = tr1d(values, dcm)
                values = values.reshape(nE, nRHS, nP, nDOF)

        values = np.moveaxis(values, 1, -1)  # (nE, nP, nDOF, nRHS)

        if flatten:
            nE, nP, nDOF, nRHS = values.shape
            values = values.reshape(nE, nP * nDOF, nRHS)

        # values -> (nE, nP, nDOF, nRHS) or (nE, nP * nDOF, nRHS)
        return values

    @squeeze(True)
    def strains(self, *args, cells: Union[int, Iterable[int]] = None, rng: Iterable = None,
                points: Union[float, Iterable] = None, **kwargs) -> ndarray:
        """
        Returns strains for one or more cells.

        Parameters
        ----------
        points : float or Iterable[float], Optional
                Points of evaluation. If provided, it is assumed that the given values
                are wrt. the range [0, 1], unless specified otherwise with the 'rng' 
                parameter. If not provided, results are returned for the nodes of the 
                selected elements. Default is None.

        rng : Iterable, Optional
            Range where the points of evauation are understood. Only for 1d cells.
            Default is [0, 1].

        cells : int or Iterable[int], Optional
            Indices of cells. If not provided, results are returned for all cells.
            Default is None.

        Returns
        -------
        numpy.ndarray
            An array of shape (nE, nP, nSTRE, nRHS). 

        """
        dofsol = self.dof_solution(squeeze=False, flatten=True, cells=cells)
        # (nE, nEVAB, nRHS)

        if cells is not None:
            cells = atleast1d(cells)
            conds = np.isin(cells, self.id)
            cells = atleast1d(cells[conds])
            assert len(cells) > 0, "Length of cells is zero!"
        else:
            cells = np.s_[:]

        if points is None:
            points = np.array(self.lcoords()).flatten()
            rng = [-1, 1]
        else:
            rng = np.array([0, 1]) if rng is None else np.array(rng)
        points, rng = to_range_1d(
            points, source=rng, target=[-1, 1]).flatten(), [-1, 1]

        dshp = self.shape_function_derivatives(points, rng=rng)[cells]
        ecoords = self.local_coordinates()[cells]
        jac = self.jacobian_matrix(dshp=dshp, ecoords=ecoords)
        # (nE, nP, nD, nD)
        B = self.strain_displacement_matrix(dshp=dshp, jac=jac)
        # (nE, nP, nSTRE, nEVAB)
        dofsol = ascont(np.swapaxes(dofsol, 1, 2))  # (nE, nRHS, nEVAB)
        strains = approx_element_solution_bulk(
            dofsol, B)  # (nE, nRHS, nP, nSTRE)
        strains = ascont(np.moveaxis(strains, 1, -1))   # (nE, nP, nSTRE, nRHS)
        return strains

    @squeeze(True)
    def kinetic_strains(self, *args, cells: Union[int, Iterable[int]] = None,
                        points: Union[float, Iterable] = None, **kwargs) -> ndarray:
        """
        Returns kinetic strains for one or more cells.

        Parameters
        ----------
        points : float or Iterable, Optional
                 Points of evaluation. If provided, it is assumed that the given values
                 are wrt. the range [0, 1], unless specified otherwise with the 'rng' 
                 parameter. If not provided, results are returned for the nodes of the 
                 selected elements. Default is None.

        rng : Iterable, Optional
            Range where the points of evauation are understood. Default is [0, 1].

        cells : int or Iterable[int], Optional
            Indices of cells. If not provided, results are returned for all cells.
            Default is None.

        Returns
        -------
        numpy.ndarray 
            An array of shape (nE, nSTRE, nRHS), where nE is the number of cells,
            nSTRE is the number of strain components and nRHS is the number of 
            load cases.

        """
        key = self._dbkey_strain_loads_
        try:
            sloads = self.db[key].to_numpy()
        except Exception as e:
            if key not in self.db.fields:
                nE = len(self)
                nodal_loads = self.pointdata.loads
                if len(nodal_loads.shape) == 2:
                    nRHS = 1
                else:
                    nRHS = self.pointdata.loads.shape[-1]
                nSTRE = self.__class__.NSTRE
                sloads = np.zeros((nE, nSTRE, nRHS))
            else:
                raise e
        finally:
            sloads = atleastnd(sloads, 3, back=True)  # (nE, nSTRE=4, nRHS)

        if cells is not None:
            cells = atleast1d(cells)
            conds = np.isin(cells, self.id)
            cells = atleast1d(cells[conds])
            if len(cells) == 0:
                return {}
        else:
            cells = np.s_[:]

        if isinstance(points, Iterable):
            nP = len(points)
            return explode_kinetic_strains(sloads[cells], nP)
        else:
            return sloads[cells]

    @squeeze(True)
    def external_forces(self, *args, cells: Union[int, Iterable[int]] = None,
                        target: Union[str, ReferenceFrame] = 'local',
                        flatten: bool = True, **kwargs) -> ndarray:
        """
        Evaluates :math:`\mathbf{f}_e = \mathbf{K}_e @ \mathbf{u}_e` for one 
        or several cells and load cases.

        Parameters
        ----------
        cells : int or Iterable[int], Optional
            Indices of cells. If not provided, results are returned for all cells.
            Default is None.

        target : Union[str, ReferenceFrame], Optional
            The target frame. Default is 'local', which means that the returned 
            forces should be understood as coordinates of generalized vectors in 
            the local frames of the cells.

        flatten: bool, Optional
            Determines the shape of the resulting array. Default is True.

        Returns
        -------
        numpy.ndarray
            An array of shape (nE, nP * nDOF, nRHS) if 'flatten' is True else 
            (nE, nP, nDOF, nRHS).

        """
        dofsol = self.dof_solution(squeeze=False, flatten=True, cells=cells)
        # (nE, nNE * nDOF, nRHS)

        if cells is not None:
            cells = atleast1d(cells)
            conds = np.isin(cells, self.id)
            cells = atleast1d(cells[conds])
            assert len(cells) > 0, "Length of cells is zero!"
        else:
            cells = np.s_[:]

        # approximation matrix
        # values -> (nE, nEVAB, nRHS)
        points = np.array(self.lcoords()).flatten()
        N = self.shape_function_matrix(points, rng=[-1, 1])[cells]
        # N -> (nE, nNE, nDOF, nDOF * nNODE)

        # calculate external forces
        dbkey = self._dbkey_stiffness_matrix_
        K = self.db[dbkey].to_numpy()
        dofsol = ascont(np.swapaxes(dofsol, 1, 2))  # (nE, nRHS, nEVAB)
        forces = calculate_external_forces_bulk(K, dofsol)
        # forces -> (nE, nRHS, nEVAB)
        forces = approx_element_solution_bulk(forces, N)
        # forces -> (nE, nRHS, nNE, nDOF)
        dofsol = ascont(np.swapaxes(dofsol, 1, 2))  # (nE, nEVAB, nRHS)
        forces = ascont(np.moveaxis(forces, 1, -1))
        # forces ->  (nE, nNE, nDOF, nRHS)

        if target is not None:
            if isinstance(target, str) and target == 'local':
                values = forces
            else:
                # transform values to a destination frame, otherwise return
                # the forces are in the local frames of the cells
                values = np.moveaxis(forces, -1, 1)
                nE, nRHS, nP, nDOF = values.shape
                values = values.reshape(nE, nRHS, nP * nDOF)
                dcm = self.direction_cosine_matrix(N=nP, target=target)[cells]
                values = tr1d(values, dcm)
                values = values.reshape(nE, nRHS, nP, nDOF)
                values = np.moveaxis(forces, 1, -1)
        else:
            values = forces

        if flatten:
            nE, nP, nX, nRHS = values.shape
            values = values.reshape(nE, nP * nX, nRHS)

        # forces : (nE, nP, nDOF, nRHS)
        return values

    @squeeze(True)
    def internal_forces(self, *args, cells: Union[int, Iterable[int]] = None,
                        rng: Iterable = None, points: Union[float, Iterable] = None,
                        flatten: bool = True, target: Union[str, ReferenceFrame] = 'local',
                        **kwargs) -> ndarray:
        """
        Returns internal forces for many cells and evaluation points.

        Parameters
        ----------
        points : float or Iterable[float], Optional
                Points of evaluation. If provided, it is assumed that the given values
                are wrt. the range [0, 1], unless specified otherwise with the 'rng' 
                parameter. If not provided, results are returned for the nodes of the 
                selected elements. Default is None.

        rng : Iterable[float], Optional
            Range where the points of evauation are understood. 
            Only for 1d cells. Default is [0, 1].

        cells : int or Iterable[int], Optional
            Indices of cells. If not provided, results are returned for all cells.
            Default is None.

        target : Union[str, ReferenceFrame], Optional
            The target frame. Default is 'local', which means that the returned forces 
            should be understood as coordinates of generalized vectors in the local 
            frames of the cells.

        flatten: bool, Optional
            Determines the shape of the resulting array. Default is True.

        Returns
        -------
        numpy.ndarray
            An array of shape (nE, nP * nDOF, nRHS) if 'flatten' is True else 
            (nE, nP, nDOF, nRHS).

        """
        dofsol = self.dof_solution(squeeze=False, flatten=True, cells=cells)
        # (nE, nNE * nDOF, nRHS)

        if cells is not None:
            cells = atleast1d(cells)
            conds = np.isin(cells, self.id)
            cells = atleast1d(cells[conds])
            assert len(cells) > 0, "Length of cells is zero!"
        else:
            cells = np.s_[:]

        if points is None:
            points = np.array(self.lcoords()).flatten()
            rng = [-1, 1]
        else:
            if isinstance(points, Iterable):
                points = np.array(points)
            rng = np.array([0, 1]) if rng is None else np.array(rng)

        # approximate at points
        # values : (nE, nEVAB, nRHS)
        points, rng = to_range_1d(
            points, source=rng, target=[-1, 1]).flatten(), [-1, 1]

        dshp = self.shape_function_derivatives(points, rng=rng)[cells]
        ecoords = self.local_coordinates()[cells]
        jac = self.jacobian_matrix(dshp=dshp, ecoords=ecoords)
        # jac -> (nE, nP, 1, 1)
        B = self.strain_displacement_matrix(dshp=dshp, jac=jac)
        # B -> (nE, nP, nSTRE, nNODE * 6)

        dofsol = ascont(np.swapaxes(dofsol, 1, 2))  # (nE, nRHS, nEVAB)
        strains = approx_element_solution_bulk(dofsol, B)
        # -> (nE, nRHS, nP, nSTRE)
        strains -= self.kinetic_strains(points=points,
                                        squeeze=False)[cells]
        D = self.model_stiffness_matrix()[cells]
        forces = calculate_internal_forces_bulk(strains, D)
        # forces -> (nE, nRHS, nP, nSTRE)
        # (nE, nP, nSTRE, nRHS)
        forces = ascont(np.moveaxis(forces, 1, -1))
        dofsol = ascont(np.swapaxes(dofsol, 1, 2))  # (nE, nEVAB, nRHS)
        # forces -> (nE, nP, nSTRE, nRHS)
        forces = self._postproc_local_internal_forces_(forces, points=points, rng=rng,
                                                       cells=cells, dofsol=dofsol)
        # forces -> (nE, nP, nDOF, nRHS)

        if target is not None:
            if isinstance(target, str) and target == 'local':
                values = forces
            else:
                # transform values to a destination frame, otherwise return
                # the forces are in the local frames of the cells
                values = np.moveaxis(forces, -1, 1)
                nE, nRHS, nP, nDOF = values.shape
                values = values.reshape(nE, nRHS, nP * nDOF)
                dcm = self.direction_cosine_matrix(N=nP, target=target)[cells]
                values = tr1d(values, dcm)
                values = values.reshape(nE, nRHS, nP, nDOF)
                values = np.moveaxis(values, 1, -1)
        else:
            values = forces

        if flatten:
            nE, nP, nSTRE, nRHS = values.shape
            values = values.reshape(nE, nP * nSTRE, nRHS)

        return values

    def global_dof_numbering(self, *args, **kwargs) -> Union[TopologyArray, ndarray]:
        """
        Returns global numbering of the degrees of freedom of the cells.        
        """
        topo = kwargs.get("topo", None)
        topo = self.topology() if topo is None else topo
        nDOFN = self.container.NDOFN
        try:
            if topo.is_jagged():
                cuts = topo.widths() * nDOFN
                data1d = np.zeros(np.sum(cuts), dtype=int)
                gnum = JaggedArray(data1d, cuts=cuts)
                topo_to_gnum_jagged(topo, gnum, nDOFN)
                return TopologyArray(gnum, to_numpy=False)
            else:
                return topo_to_gnum(topo.to_numpy(), nDOFN)
        except Exception:
            return topo_to_gnum(topo, nDOFN)

    @squeeze(True)
    def elastic_stiffness_matrix(self, *, transform: bool = True,
                                 minval: float = 1e-12, sparse: bool = False,
                                 **kwargs) -> Union[ndarray, coo_matrix]:
        """
        Returns the elastic stiffness matrix of the cells.

        Parameters
        ----------
        transform : bool, Optional
            If True, local matrices are transformed to the global frame.
            Default is True.

        minval : float, Optional
            A minimal value for the entries in the main diagonal. Set it to a negative
            value to diable its effect. Default is 1e-12.

        sparse : bool, Optional
            If True, the returned object is a sparse COO matrix. Default is False.

        Returns
        -------
        numpy.ndarray or scipy.sparse.coo_matrix
            A sparse SciPy matrix if 'sparse' is True, or a 3d numpy array, where the 
            elements run along the first axis.

        """
        dbkey = self._dbkey_stiffness_matrix_
        if dbkey not in self.db.fields:
            K = self._elastic_stiffness_matrix_(transform=False, **kwargs)
            assert_min_diagonals_bulk(K, minval)
            self.db[dbkey] = K
        else:
            K = self.db[dbkey].to_numpy()
        # if the model has more dofs than the element
        nDOFN = self.container.NDOFN
        dofmap = self.__class__.dofmap
        if len(dofmap) < nDOFN:
            nE = K.shape[0]
            nX = nDOFN * self.NNODE
            K_ = np.zeros((nE, nX, nX), dtype=float)
            dofmap = element_dofmap_bulk(dofmap, nDOFN, self.NNODE)
            K = expand_coeff_matrix_bulk(K, K_, dofmap)
            assert_min_diagonals_bulk(K, minval)
        if transform:
            K = self._transform_coeff_matrix_(K)
        if sparse:
            assert transform, "Must be transformed for a sparse result."
            nP = len(self.pointdata)
            N = nP * self.NDOFN
            topo = self.topology().to_numpy()
            gnum = self.global_dof_numbering(topo=topo)
            K = fem_coeff_matrix_coo(K, inds=gnum, N=N, **kwargs)
        return K

    @config(store_strains=False)
    def _elastic_stiffness_matrix_(self, *args, transform: bool = True, **kwargs):
        ec = kwargs.get('_ec', None)
        if ec is None:
            nSTRE = self.__class__.NSTRE
            nDOF = self.__class__.NDOFN
            nNE = self.__class__.NNODE
            nE = len(self)
            _topo = kwargs.get('_topo', self.topology().to_numpy())
            _frames = kwargs.get('_frames', self.frames)
            if _frames is not None:
                ec = self.local_coordinates(_topo=_topo, frames=_frames)
            else:
                ec = self.points_of_cells(topo=_topo)
            dbkey = self._dbkey_strain_displacement_matrix_
            self.db[dbkey] = np.zeros((nE, nSTRE, nDOF * nNE))

        q = kwargs.get('_q', None)
        if q is None:
            D = self.model_stiffness_matrix()
            func = partial(self._elastic_stiffness_matrix_, _D=D, **kwargs)
            q = self.quadrature[self.qrule]
            N = nDOF * nNE
            K = np.zeros((len(self), N, N))
            if isinstance(q, dict):
                for qinds, qvalue in q.items():
                    if isinstance(qvalue, str):
                        qpos, qweight = self.quadrature[qvalue]
                    else:
                        qpos, qweight = qvalue
                    q = Quadrature(qinds, qpos, qweight)
                    K += func(_q=q, _ec=ec)
            else:
                qpos, qweight = self.quadrature[self.qrule]
                q = Quadrature(None, qpos, qweight)
                K = func(_q=q, _ec=ec)
            dbkey = self._dbkey_stiffness_matrix_
            self.db[dbkey] = K
            return self._transform_coeff_matrix_(K) if transform else K

        # in side loop
        dshp = self.shape_function_derivatives(q.pos)
        jac = self.jacobian_matrix(dshp=dshp, ecoords=ec)
        B = self.strain_displacement_matrix(dshp=dshp, jac=jac)
        if q.inds is not None:
            # zero out unused indices, only for selective integration
            inds = np.where(~np.in1d(np.arange(self.NSTRE), q.inds))[0]
            B[:, :, inds, :] = 0.
        djac = self.jacobian(jac=jac)
        # B (nE, nG, nSTRE=4, nNODE * nDOF=6)
        dbkey = self._dbkey_strain_displacement_matrix_
        _B = self.db[dbkey].to_numpy()
        _B += strain_displacement_matrix_bulk2(B, djac, q.weight)
        self.db[dbkey] = _B
        D = kwargs.get('_D', self.model_stiffness_matrix())
        return stiffness_matrix_bulk2(D, B, djac, q.weight)

    @squeeze(True)
    def consistent_mass_matrix(self, *args, sparse: bool = False,
                               transform: bool = True, minval: float = 1e-12,
                               **kwargs) -> Union[ndarray, coo_matrix]:
        """
        Returns the stiffness-consistent mass matrix of the cells.

        Parameters
        ----------
        transform : bool, Optional
            If True, local matrices are transformed to the global frame.
            Default is True.

        minval : float, Optional
            A minimal value for the entries in the main diagonal. Set it to a 
            negative value to diable its effect. Default is 1e-12.

        sparse : bool, Optional
            If True, the returned object is a sparse COO matrix. 
            Default is False.

        Returns
        -------
        numpy.ndarray or scipy.sparse.coo_matrix
            A sparse SciPy matrix if 'sparse' is True, or a 3d numpy array, 
            where the elements run along the first axis.

        """
        dbkey = self._dbkey_mass_matrix_
        if dbkey not in self.db.fields:
            M = self._consistent_mass_matrix_(transform=False, **kwargs)
            assert_min_diagonals_bulk(M, minval)
            self.db[dbkey] = M
        else:
            M = self.db[dbkey].to_numpy()
        # if the model has more dofs than the element
        nDOFN = self.container.NDOFN
        dofmap = self.__class__.dofmap
        if len(dofmap) < nDOFN:
            nE = M.shape[0]
            nX = nDOFN * self.NNODE
            M_ = np.zeros((nE, nX, nX), dtype=float)
            dofmap = element_dofmap_bulk(dofmap, nDOFN, self.NNODE)
            M = expand_coeff_matrix_bulk(M, M_, dofmap)
            assert_min_diagonals_bulk(M, minval)
        if transform:
            M = self._transform_coeff_matrix_(M)
        if sparse:
            nP = len(self.pointdata)
            N = nP * self.NDOFN
            topo = self.topology().to_numpy()
            gnum = self.global_dof_numbering(topo=topo)
            M = fem_coeff_matrix_coo(M, *args, inds=gnum, N=N, **kwargs)
        return M

    def _consistent_mass_matrix_(self, *args, values=None,
                                 transform: bool = True, **kwargs):
        _topo = kwargs.get('_topo', self.topology().to_numpy())
        if 'frames' not in kwargs:
            key = self._dbkey_frames_
            if key in self.db.fields:
                frames = self.frames
            else:
                frames = None
        _ecoords = kwargs.get('_ecoords', None)
        if _ecoords is None:
            if frames is not None:
                _ecoords = self.local_coordinates(_topo=_topo, frames=frames)
            else:
                _ecoords = self.points_of_cells(topo=_topo)
        if isinstance(values, np.ndarray):
            _dens = values
        else:
            _dens = kwargs.get('_dens', self.density)

        try:
            _areas = self.areas()
        except Exception:
            _areas = np.ones_like(_dens)

        q = kwargs.get('_q', None)
        if q is None:
            q = self.quadrature['mass']
            func = partial(self._consistent_mass_matrix_, **kwargs)
            if isinstance(q, dict):
                N = self.NNODE * self.NDOFN
                nE = len(self)
                res = np.zeros((nE, N, N))
                for qinds, qvalue in q.items():
                    if isinstance(qvalue, str):
                        qpos, qweight = self.quadrature[qvalue]
                    else:
                        qpos, qweight = qvalue
                    q = Quadrature(qinds, qpos, qweight)
                    res += func(*args, _topo=_topo, frames=frames, _q=q,
                                values=_dens, _ecoords=_ecoords)
            else:
                qpos, qweight = self.quadrature['mass']
                q = Quadrature(None, qpos, qweight)
                res = func(*args, _topo=_topo, frames=frames, _q=q,
                           values=_dens, _ecoords=_ecoords)
            key = self._dbkey_mass_matrix_
            self.db[key] = res
            return self._transform_coeff_matrix_(res) if transform else res

        rng = np.array([-1., 1.])
        dshp = self.shape_function_derivatives(q.pos)
        jac = self.jacobian_matrix(dshp=dshp, ecoords=_ecoords)
        djac = self.jacobian(jac=jac)
        N = self.shape_function_matrix(q.pos, rng=rng)
        return mass_matrix_bulk(N, _dens, _areas, djac, q.weight)

    @squeeze(True)
    def load_vector(self, transform: bool = True, assemble: bool = False,
                    **kwargs) -> ndarray:
        """
        Builds the equivalent nodal load vector from all sources
        and returns it in either the global frame or cell-local frames.

        Parameters
        ----------
        assemble : bool, Optional
            If True, the values are returned with a matching shape to the total
            system. Default is False.

        transform : bool, Optional
            If True, local matrices are transformed to the global frame.
            Default is True.

        See Also
        --------
        :func:`body_load_vector`
        :func:`strain_load_vector`

        Returns
        -------
        numpy.ndarray
            The nodal load vector for all load cases as a 2d numpy array
            of shape (nX, nRHS), where nX and nRHS are the total number of 
            unknowns of the structure and the number of load cases.

        """
        dbkey = self._dbkey_nodal_load_vector_
        if dbkey not in self.db.fields:
            options = dict(transform=False, assemble=False,
                           squeeze=False, return_zeroes=True)
            f = self.body_load_vector(**options)
            f += self.strain_load_vector(**options)
            self.db[dbkey] = f
        else:
            f = self.db[dbkey].to_numpy()
        if transform:
            nodal_loads = self._transform_nodal_loads_(nodal_loads)
            # (nE, nRHS, nNE * nDOF) -> (nE, nNE * nDOF, nRHS)
        if assemble:
            assert transform, "Must transform before assembly."
            nodal_loads = self._assemble_nodal_loads_(nodal_loads)
            # (nX, nRHS)
        return f

    @squeeze(True)
    def strain_load_vector(self, values: ndarray = None, *,
                           return_zeroes: bool = False, transform: bool = True,
                           assemble: bool = False, **kwargs) -> ndarray:
        """
        Generates a load vector from strain loads specified for all cells.

        Parameters
        ----------
        values : numpy.ndarray, Optional
            Strain loads as a 3d numpy array of shape (nE, nS, nRHS). 
            The array must contain values for all cells (nE), strain 
            components (nS) and load cases (nL).

        return_zeroes : bool, Optional
            Controls what happends if there are no strain loads provided.
            If True, a zero array is retured with correct shape, otherwise None. 
            Default is False.

        transform : bool, Optional
            If True, local matrices are transformed to the global frame.
            Default is True.

        assemble : bool, Optional
            If True, the values are returned with a matching shape to the total
            system. Default is False.

        See Also
        --------
        :func:`load_vector`
        :func:`body_load_vector`

        Returns
        -------
        numpy.ndarray
            The equivalent load vector.

        """
        dbkey = self._dbkey_strain_loads_
        try:
            if values is None:
                values = self.db[dbkey].to_numpy()
        except Exception as e:
            if dbkey not in self.db.fields:
                if not return_zeroes:
                    return None
                if len(self.pointdata.loads.shape) == 2:
                    nRHS = 1
                else:
                    nRHS = self.pointdata.loads.shape[-1]
                nE = len(self)
                nSTRE = self.__class__.NSTRE
                values = np.zeros((nE, nSTRE, nRHS))
            else:
                raise e
        finally:
            values = atleastnd(values, 3, back=True)  # (nE, nSTRE=4, nRHS)

        dbkey = self._dbkey_strain_displacement_matrix_
        if dbkey not in self.db.fields:
            self.elastic_stiffness_matrix(squeeze=False, transform=False)
        B = self.db[dbkey].to_numpy()  # (nE, nSTRE=4, nNODE * nDOF=6)
        D = self.model_stiffness_matrix()  # (nE, nSTRE=4, nSTRE=4)
        BTD = unit_strain_load_vector_bulk(D, B)
        values = np.swapaxes(values, 1, 2)  # (nE, nRHS, nSTRE=4)
        nodal_loads = strain_load_vector_bulk(BTD, ascont(values))

        if transform:
            nodal_loads = self._transform_nodal_loads_(nodal_loads)
            # (nE, nRHS, nNE * nDOF) -> (nE, nNE * nDOF, nRHS)

        if assemble:
            assert transform, "Must transform before assembly."
            nodal_loads = self._assemble_nodal_loads_(nodal_loads)
            # (nX, nRHS)

        return nodal_loads

    @squeeze(True)
    def body_load_vector(self, values: ndarray = None, *, constant: bool = False,
                         return_zeroes: bool = False, transform: bool = True,
                         assemble: bool = False, **kwargs) -> ndarray:
        """
        Builds the equivalent discrete representation of body loads
        and returns it in either the global frame or cell-local frames.

        Parameters
        ----------
        values : numpy.ndarray, Optional
            Body load values for all cells. Default is None.

        constant : bool, Optional
            Set this True if the input represents a constant load.
            Default is False.

        assemble : bool, Optional
            If True, the values are returned with a matching shape to the total
            system. Default is False.

        return_zeroes : bool, Optional
            Controls what happends if there are no strain loads provided.
            If True, a zero array is retured with correct shape, otherwise None. 
            Default is False.

        transform : bool, Optional
            If True, local matrices are transformed to the global frame.
            Default is True.

        See Also
        --------
        :func:`load_vector`
        :func:`strain_load_vector`

        Returns
        -------
        numpy.ndarray
            The nodal load vector for all load cases as a 2d numpy array
            of shape (nX, nRHS), where nX and nRHS are the total number of 
            unknowns of the structure and the number of load cases.

        """
        nNE = self.__class__.NNODE
        dbkey = self._dbkey_body_loads_
        try:
            if values is None:
                values = self.db[dbkey].to_numpy()
        except Exception as e:
            if dbkey not in self.db.fields:
                if not return_zeroes:
                    return None
                if len(self.pointdata.loads.shape) == 2:
                    nRHS = 1
                else:
                    nRHS = self.pointdata.loads.shape[-1]
                nE = len(self)
                nDOF = self.__class__.NDOFN
                values = np.zeros((nE, nNE, nDOF, nRHS))
            else:
                raise e
        finally:
            values = atleastnd(values, 4, back=True)  # (nE, nNE, nDOF, nRHS)

        # prepare data to shape (nE, nNE * nDOF, nRHS)
        if constant:
            values = atleastnd(values, 3, back=True)  # (nE, nDOF, nRHS)
            #np.insert(values, 1, values, axis=1)
            nE, nDOF, nRHS = values.shape
            values_ = np.zeros((nE, nNE, nDOF, nRHS), dtype=values.dtype)
            for i in range(nNE):
                values_[:, i, :, :] = values
            values = values_
        values = atleastnd(values, 4, back=True)  # (nE, nNE, nDOF, nRHS)
        nE, _, nDOF, nRHS = values.shape
        # (nE, nNE, nDOF, nRHS) -> (nE, nNE * nDOF, nRHS)
        values = values.reshape(nE, nNE * nDOF, nRHS)
        values = ascont(values)
        nodal_loads = self.integrate_body_loads(values)
        # (nE, nNE * nDOF, nRHS)

        if transform:
            nodal_loads = self._transform_nodal_loads_(nodal_loads)
            # (nE, nRHS, nNE * nDOF) -> (nE, nNE * nDOF, nRHS)

        if assemble:
            assert transform, "Must transform before assembly."
            nodal_loads = self._assemble_nodal_loads_(nodal_loads)
            # (nX, nRHS)

        return nodal_loads

    def condensate(self):
        """
        Applies static condensation to account for cell fixity.

        References
        ----------
        .. [1] Duan Jin, Li-Yun-gui "About the Finite Element 
           Analysis for Beam-Hinged Frame," Advances in Engineering 
           Research, vol. 143, pp. 231-235, 2017.

        """
        if not self.has_fixity:
            return
        else:
            fixity = self.fixity
            nE, nNE, nDOF = fixity.shape
            fixity = fixity.reshape(nE, nNE * nDOF)
        dbkey_K = self._dbkey_stiffness_matrix_
        dbkey_M = self._dbkey_mass_matrix_
        dbkey_f = self._dbkey_nodal_load_vector_
        K = self.db[dbkey_K].to_numpy()
        f = self.db[dbkey_f].to_numpy()
        nEVAB_full = nNE * nDOF - 0.001
        cond = np.sum(fixity, axis=1) < nEVAB_full
        i = np.where(cond)[0]
        K[i], f[i] = condensate_Kf_bulk(K[i], f[i], fixity[i])
        self.db[dbkey_K] = K
        self.db[dbkey_f] = f
        if dbkey_M in self.db.fields:
            M = self.db[dbkey_M].to_numpy()
            M[i] = condensate_M_bulk(M[i], fixity[i])
            self.db[dbkey_M] = M

    def _transform_coeff_matrix_(self, A: ndarray, *args, invert: bool = False,
                                 **kwargs) -> ndarray:
        """
        Transforms element coefficient matrices (eg. the stiffness or the 
        mass matrix) from local to global.

        Parameters
        ----------
        *args
            Forwarded to :func:`direction_cosine_matrix`

        **kwargs
            Forwarded to :func:`direction_cosine_matrix`

        A : numpy.ndarray
            The coefficient matrix in the source frame.

        invert : bool, Optional
            If True, the DCM matrices are transposed before transformation.
            This makes this function usable in both directions.
            Default is False.

        Returns
        -------
        numpy.ndarray
            The coefficient matrix in the target frame.

        """
        # DCM from local to global
        dcm = self.direction_cosine_matrix(*args, **kwargs)
        return A if dcm is None else tr2d(A, dcm, invert)

    def _transform_nodal_loads_(self, nodal_loads: ndarray) -> ndarray:
        """
        Transforms discrete nodal loads to the global frame.

        Parameters
        ----------
        nodal_loads : numpy.ndarray
            A 3d array of shape (nE, nEVAB, nRHS).

        Returns
        -------
        numpy.ndarray
            A numpy array of shape (nE, nEVAB, nRHS).

        """
        dcm = self.direction_cosine_matrix(target='global')
        # (nE, nNE * nDOF, nRHS) -> (nE, nRHS, nNE * nDOF)
        nodal_loads = np.swapaxes(nodal_loads, 1, 2)
        nodal_loads = ascont(nodal_loads)
        nodal_loads = tr1d(nodal_loads, dcm)
        nodal_loads = np.swapaxes(nodal_loads, 1, 2)
        # (nE, nRHS, nNE * nDOF) -> (nE, nNE * nDOF, nRHS)
        return nodal_loads

    def _assemble_nodal_loads_(self, nodal_loads: ndarray) -> ndarray:
        """ 
        Assembles the nodal load vector for multiple load cases.

        Parameters
        ----------
        nodal_loads : numpy.ndarray
            A 3d array of shape (nE, nEVAB, nRHS).

        Returns
        -------
        numpy.ndarray
            A numpy array of shape (nX, nRHS), where nX is the total
            number of unknowns in the total structure.

        """
        topo = self.topology().to_numpy()
        gnum = self.global_dof_numbering(topo=topo)
        nX = len(self.pointdata) * self.container.NDOFN
        return assemble_load_vector(nodal_loads, gnum, nX)  # (nX, nRHS)



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from numba import njit, prange
__cache = True


@njit(nogil=True, cache=__cache)
def shape_function_values(x: float, L: float = 2):
    """
    Evaluates the shape functions at a point x in the range [-1, 1].

    Parameters
    ----------
    x : float
        A point in the range [-1, 1].

    L : float, Optional
        Length of the beam element. Default is 2.

    Returns
    -------
    numpy.ndarray
        A 2d float array.

    """
    return np.array([
        [
            0.5 * x * (x - 1),
            x**2 * (0.75 * x + 1.0) * (x - 1)**2,
            x**2 * (0.75 * x + 1.0) * (x - 1)**2,
            0.5 * x * (x - 1),
            -0.125 * L * x**2 * (x - 1)**2 * (x + 1),
            0.125 * L * x**2 * (x - 1)**2 * (x + 1)
        ],
        [
            1.0 - 1.0 * x**2,
            1.0 * (x - 1)**2 * (x + 1)**2,
            1.0 * (x - 1)**2 * (x + 1)**2,
            1.0 - 1.0 * x**2,
            L * x * (-0.5 * x**4 + 1.0 * x**2 - 0.5),
            0.5 * L * x * (x - 1)**2 * (x + 1)**2
        ],
        [
            0.5 * x * (x + 1),
            x**2 * (1.0 - 0.75 * x) * (x + 1)**2,
            x**2 * (1.0 - 0.75 * x) * (x + 1)**2,
            0.5 * x * (x + 1),
            -0.125 * L * x**2 * (x - 1) * (x + 1)**2,
            0.125 * L * x**2 * (x - 1) * (x + 1)**2
        ]
    ])


@njit(nogil=True, cache=__cache)
def shape_function_derivatives_1(x: float, L: float = 2):
    """
    Evaluates the first derivatives of the shape
    functions at a point x in the range [-1, 1].

    Parameters
    ----------
    x : float
        A point in the range [-1, 1].

    L : float, Optional
        Length of the beam element. Default is 2.

    Returns
    -------
    numpy.ndarray
        A 2d float array.

    """
    return np.array([
        [
            1.0 * x - 0.5,
            3.75 * x * (x - 1) * (1.0 * x - 0.533333333333333) * (x + 1),
            3.75 * x * (x - 1) * (1.0 * x - 0.533333333333333) * (x + 1),
            1.0 * x - 0.5,
            L * x * (-0.625 * x**3 + 0.5 * x**2 + 0.375 * x - 0.25),
            0.625 * L * x * (x - 1) * (1.0 * x**2 + 0.2 * x - 0.4)
        ],
        [
            -2.0 * x,
            4.0 * x * (x**2 - 1),
            4.0 * x * (x**2 - 1),
            -2.0 * x,
            L * (-2.5 * x**4 + 3.0 * x**2 - 0.5),
            L * (2.5 * x**4 - 3.0 * x**2 + 0.5)
        ],
        [
            1.0 * x + 0.5,
            -3.75 * x * (x - 1) * (1.0 * x + 0.533333333333333) * (x + 1),
            -3.75 * x * (x - 1) * (1.0 * x + 0.533333333333333) * (x + 1),
            1.0 * x + 0.5,
            0.625 * L * x * (x + 1) * (-1.0 * x**2 + 0.2 * x + 0.4),
            L * x * (0.625 * x**3 + 0.5 * x**2 - 0.375 * x - 0.25)
        ]
    ])


@njit(nogil=True, cache=__cache)
def shape_function_derivatives_2(x: float, L: float = 2):
    """
    Evaluates the second derivatives of the shape
    functions at a point x in the range [-1, 1].

    Parameters
    ----------
    x : float
        A point in the range [-1, 1].

    L : float, Optional
        Length of the beam element. Default is 2.

    Returns
    -------
    numpy.ndarray
        A 2d float array.

    """
    return np.array([
        [
            1.00000000000000,
            15.0 * x**3 - 6.0 * x**2 - 7.5 * x + 2.0,
            15.0 * x**3 - 6.0 * x**2 - 7.5 * x + 2.0,
            1.00000000000000,
            L * (-2.5 * x**3 + 1.5 * x**2 + 0.75 * x - 0.25),
            L * (2.5 * x**3 - 1.5 * x**2 - 0.75 * x + 0.25)
        ],
        [
            -2.00000000000000,
            12.0 * x**2 - 4.0,
            12.0 * x**2 - 4.0,
            -2.00000000000000,
            L * x * (6.0 - 10.0 * x**2),
            L * x * (10.0 * x**2 - 6.0)
        ],
        [
            1.00000000000000,
            -15.0 * x**3 - 6.0 * x**2 + 7.5 * x + 2.0,
            -15.0 * x**3 - 6.0 * x**2 + 7.5 * x + 2.0,
            1.00000000000000,
            L * (-2.5 * x**3 - 1.5 * x**2 + 0.75 * x + 0.25),
            L * (2.5 * x**3 + 1.5 * x**2 - 0.75 * x - 0.25)
        ]
    ])


@njit(nogil=True, cache=__cache)
def shape_function_derivatives_3(x: float, L: float = 2):
    """
    Evaluates the third derivatives of the shape
    functions at a point x in the range [-1, 1].

    Parameters
    ----------
    x : float
        A point in the range [-1, 1].

    L : float, Optional
        Length of the beam element. Default is 2.

    Returns
    -------
    numpy.ndarray
        A 2d float array.

    """
    return np.array([
        [
            0,
            45.0 * x**2 - 12.0 * x - 7.5,
            45.0 * x**2 - 12.0 * x - 7.5,
            0,
            L * (-7.5 * x**2 + 3.0 * x + 0.75),
            L * (7.5 * x**2 - 3.0 * x - 0.75)
        ],
        [
            0,
            24.0 * x,
            24.0 * x,
            0,
            L * (6.0 - 30.0 * x**2),
            L * (30.0 * x**2 - 6.0)
        ],
        [
            0,
            -45.0 * x**2 - 12.0 * x + 7.5,
            -45.0 * x**2 - 12.0 * x + 7.5,
            0,
            L * (-7.5 * x**2 - 3.0 * x + 0.75),
            L * (7.5 * x**2 + 3.0 * x - 0.75)
        ]
    ])


@njit(nogil=True, parallel=True, cache=__cache)
def shape_function_values_bulk(x: ndarray, L: ndarray):
    """
    Evaluates the shape functions at several points
    in the range [-1, 1].

    Parameters
    ----------
    x : numpy.ndarray
        1d array of floats in the range [-1, -1].

    L : numpy.ndarray
        Lengths of the beam elements. Default is None.

    Returns
    -------
    numpy.ndarray
        3d float array of shape (nP, nNE=3, nDOF=6).
    """
    nP = x.shape[0]
    nE = L.shape[0]
    res = np.zeros((nE, nP, 3, 6), dtype=x.dtype)
    for iE in prange(nE):
        for iP in prange(nP):
            res[iE, iP] = shape_function_values(x[iP], L[iE])
    return res


@njit(nogil=True, cache=__cache)
def shape_function_derivatives(x: float, L: float = 2):
    """
    Evaluates the derivatives of the shape
    functions at a point x in the range [-1, 1].

    Parameters
    ----------
    x : float
        A point in the range [-1, 1].

    L : float, Optional
        Length of the beam element. Default is 2.

    Returns
    -------
    numpy.ndarray
        3d float array of shape (nNE=3, nDOF=6, 3).
    """
    res = np.zeros((3, 6, 3))
    res[:, :, 0] = shape_function_derivatives_1(x, L)
    res[:, :, 1] = shape_function_derivatives_2(x, L)
    res[:, :, 2] = shape_function_derivatives_3(x, L)
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def shape_function_derivatives_bulk(x: ndarray, L: ndarray):
    """
    Evaluates the derivatives of the shape
    functions at several points in the range [-1, 1].

    Parameters
    ----------
    x : numpy.ndarray
        1d array of floats in the range [-1, -1].

    L : numpy.ndarray
        Lengths of the beam elements. Default is None.

    Returns
    -------
    numpy.ndarray
        5d float array of shape (nE, nP, nNE=3, nDOF=6, 3).
    """
    nP = x.shape[0]
    nE = L.shape[0]
    res = np.zeros((nE, nP, 3, 6, 3), dtype=x.dtype)
    for iE in prange(nE):
        for iP in prange(nP):
            res[iE, iP] = shape_function_derivatives(x[iP], L[iE])
    return res



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from numba import njit, vectorize

__cache = True


@njit(nogil=True, cache=__cache)
def HMH_M(strs: ndarray):
    """
    Evaluates the Huber-Mises-Hencky formula for membranes.

    Parameters
    ----------
    strs : numpy.ndarray
        The stresses s11, s22, s12

    """
    s11, s22, s12 = strs
    return np.sqrt(s11**2 - s11*s22 + s22**2 + 3*s12**2)


@njit(nogil=True, cache=__cache)
def HMH_S(strs: ndarray):
    """
    Evaluates the Huber-Mises-Hencky formula for shells.

    Parameters
    ----------
    strs : numpy.ndarray
        The stresses s11, s22, s12, s13, s23

    """
    s11, s22, s12, s13, s23 = strs
    return np.sqrt(s11**2 - s11*s22 + s22**2 +
                   3*s12**2 + 3*s13**2 + 3*s23**2)


@njit(nogil=True, cache=__cache)
def HMH_3d(strs: ndarray):
    """
    Evaluates the Huber-Mises-Hencky formula for 3d solids.

    Parameters
    ----------
    strs : numpy.ndarray
        The stresses s11, s22, s33, s12, s13, s23

    """
    s11, s22, s33, s12, s13, s23 = strs
    return np.sqrt(0.5*((s11-s22)**2 + (s22-s33)**2 + (s33-s11)**2) +
                   3*(s12**2 + s13**2 + s23**2))


@vectorize("f8(f8, f8, f8, f8, f8, f8)", target='parallel', cache=__cache)
def HMH_3d_v(s11, s22, s33, s12, s13, s23):
    """
    Vectorized evaluation of the HMH failure criterion.

    The input values s11, s22, s33, s12, s13, s23 can be
    arbitrary dimensional arrays. 

    """
    return np.sqrt(0.5*((s11-s22)**2 + (s22-s33)**2 + (s33-s11)**2) +
                   3*(s12**2 + s13**2 + s23**2))



# -*- coding: utf-8 -*-
import numpy as np

from dewloosh.core.meta import ABCMeta_Weak

from .dofmap import DOF


class ABCMeta_FemMesh(ABCMeta_Weak):
    """
    Meta class for PointData and CellData classes.
    
    It merges attribute maps with those of the parent classes.
    
    """

    def __init__(self, name, bases, namespace, *args, **kwargs):
        super().__init__(name, bases, namespace, *args, **kwargs)

    def __new__(metaclass, name, bases, namespace, *args, **kwargs):
        cls = super().__new__(metaclass, name, bases, 
                              namespace, *args, **kwargs)     
        if hasattr(cls, 'dofs'):
            cls.NDOFN = len(cls.dofs)
            cls.dofmap = np.array(DOF.dofmap(cls.dofs), dtype=int)
        return cls
    
    
class ABC_FemMesh(metaclass=ABCMeta_FemMesh):
    """Helper class that provides a standard way to create an ABC using
    inheritance.
    """
    __slots__ = ()


# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from numba import njit, prange

from polymesh.cells import Q9 as Quadrilateral
from polymesh.tri.triutils import area_tri, glob_to_nat_tri, loc_to_nat_tri
from polymesh.utils import nodal_distribution_factors

from ....solid.fem.cells import FiniteElement
from ....solid.fem.model.membrane import Membrane
from ....solid.fem.utils import topo_to_gnum


__cache = True


__all__ = ['Q5MV']


topoT2 = np.array([[0, 1, 3, 4, 8, 7], [1, 2, 3, 5, 6, 8],
                   [0, 1, 2, 4, 5, 8], [0, 2, 3, 8, 6, 7]])
ndfT2 = nodal_distribution_factors(topoT2, np.ones(4))


@njit(nogil=True, parallel=True, cache=__cache)
def dofmapT2():
    """Returns the indices of dofs in each subtriangle relative
    to the index array of active nodes `activeT2`."""
    topoT2a = np.array([[0, 4, 3], [1, 2, 4],
                        [0, 1, 4], [4, 2, 3]])
    res = np.zeros((4, 6), dtype=np.int64)
    for i in prange(4):
        for j in prange(3):
            for k in prange(2):
                res[i, j*2 + k] = topoT2a[i, j]*2 + k
    return res


@njit(nogil=True, cache=__cache)
def Turner_to_Veubeke():
    """Transformation matrix of nodal variables in the 
    direction Turner -> Veubeke."""
    return np.array([[1, 0, 1, 0, 0, 0],
                     [0, 1, 0, 1, 0, 0],
                     [0, 0, 1, 0, 1, 0],
                     [0, 0, 0, 1, 0, 1],
                     [1, 0, 0, 0, 1, 0],
                     [0, 1, 0, 0, 0, 1]]) / 2


@njit(nogil=True, cache=__cache)
def shp_Veubeke_a(acoord: np.ndarray):
    A1, A2, A3 = acoord
    return np.array([A1 + A2 - A3, -A1 + A2 + A3, A1 - A2 + A3],
                    dtype=acoord.dtype)


@njit(nogil=True, cache=__cache)
def shp_Veubeke_g(gcoord: np.ndarray, ecoords: np.ndarray):
    return shp_Veubeke_a(glob_to_nat_tri(gcoord, ecoords))


@njit(nogil=True, cache=__cache)
def shp_Veubeke_p(pcoord: np.ndarray):
    return shp_Veubeke_a(loc_to_nat_tri(pcoord))


@njit(nogil=True, parallel=True, cache=__cache)
def nodal_approximation_matrix_V6():
    """Returns a matrix of nodal approximation coefficients.
    The coefficients in the ith row describe the approximation
    at node i as a linear combination of values at every node.
    At a compatible element, this is an identity matrix."""
    pcoords = np.array([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0],
                        [0.5, 0.0], [0.5, 0.5], [0.0, 0.5]])
    res = np.zeros((6, 6), dtype=pcoords.dtype)
    for i in prange(6):
        res[i, 3:] = shp_Veubeke_p(pcoords[i])
    return res


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def nodal_approximation_matrix_T2():
    """Returns a matrix of nodal approximation coefficients
    for the T2 variant, for the whole element."""
    nshpV6 = nodal_approximation_matrix_V6()
    res = np.zeros((9, 9), dtype=nshpV6.dtype)
    for iE in prange(4):
        for iN in prange(6):
            i = topoT2[iE, iN]
            ndf = ndfT2[iE, iN]
            for jN in prange(6):
                j = topoT2[iE, jN]
                res[i, j] += nshpV6[iN, jN] * ndf
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def nodal_approximation_matrix_T2_bulk(ndf: ndarray):
    """Returns a matrix of nodal approximation coefficients for all elements."""
    nE = ndf.shape[0]
    nappr = nodal_approximation_matrix_T2()
    res = np.zeros((nE, 9, 9), dtype=nappr.dtype)
    for iE in prange(nE):
        for i in prange(9):
            for j in prange(9):
                res[iE, i, j] = nappr[i, j] * ndf[iE, i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def approximation_matrix_T2_bulk(ndf: ndarray):
    """Returns a matrix of approximation coefficients for all elements."""
    nE = ndf.shape[0]
    nappr = nodal_approximation_matrix_T2()
    res = np.zeros((nE, 18, 18), dtype=nappr.dtype)
    for iE in prange(nE):
        for i in prange(9):
            for j in prange(9):
                for ii in prange(2):
                    for jj in prange(2):
                        res[iE, i*2 + ii, j*2 + jj] = nappr[i, j] * ndf[iE, i]
    return res
             

@njit(nogil=True, cache=__cache)
def _stiffness_matrix_V_(C: ndarray, ecoords: ndarray):
    (x1, x2, x3), (y1, y2, y3) = ecoords[:, 0], ecoords[:, 1]
    x12, x31, x23 = x1 - x2, x3 - x1, x2 - x3
    y21, y13, y32 = y2 - y1, y1 - y3, y3 - y2
    A = area_tri(ecoords)
    B = np.zeros((3, 6), dtype=C.dtype)
    B[0, :] = np.array([y21, 0, y32, 0, y13, 0], dtype=C.dtype)
    B[1, :] = np.array([0, x12, 0, x23, 0, x31], dtype=C.dtype)
    B[2, :] = np.array([x12, y21, x23, y32, x31, y13], dtype=C.dtype)
    return B.T @ C @ B / A


@njit(nogil=True, parallel=True, cache=__cache)
def stiffness_matrix_V(C: np.ndarray, ecoords: np.ndarray):
    nE = len(C)
    res = np.zeros((nE, 6, 6), dtype=C.dtype)
    for i in prange(nE):
        res[i, :, :] = _stiffness_matrix_V_(C[i], ecoords[i])
    return res


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def stiffness_matrix_T2(C: ndarray, ecoords: ndarray):
    nE = len(C)
    dmapT2 = dofmapT2()
    res = np.zeros((nE, 10, 10), dtype=C.dtype)
    C = C/2
    for i in prange(4):
        Ki = stiffness_matrix_V(C, ecoords[:, topoT2[i, :3], :])
        for j in prange(6):
            for k in prange(6):
                res[:, dmapT2[i, j], dmapT2[i, k]] += Ki[:, j, k]
    return res


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def transfer_passive_loads_T2(cellloadsT2: ndarray):
    """(2) Transforms loads on passive nodes to active nodes
    according to the transformation described by `Turner_to_Veubeke`."""
    nE = cellloadsT2.shape[0]
    res = np.zeros_like(cellloadsT2)
    T_to_V = Turner_to_Veubeke()
    for iE in prange(nE):
        for iSE in prange(4):
            loads_Turner = np.ravel(cellloadsT2[iE, iSE, :3, :])
            loads_Veubeke = T_to_V @ loads_Turner
            loads_Veubeke = np.reshape(loads_Veubeke, (3, 2))
            for i in prange(3, 6):
                for j in prange(2):
                    res[iE, iSE, i, j] += cellloadsT2[iE, iSE, i, j]
                    res[iE, iSE, i, j] += loads_Veubeke[i-3, j]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def distribute_nodal_data_T2(celldata: ndarray):
    """(1) Distributes master element data to subelement level."""
    nE, _, nD = celldata.shape
    res = np.zeros((nE, 4, 6, nD))
    for iE in prange(nE):
        for iSE in prange(4):
            for iSNE in prange(6):
                res[iE, iSE, iSNE] = \
                    celldata[iE, topoT2[iSE, iSNE]] * ndfT2[iSE, iSNE]
    return res


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def collect_nodal_dataT2(celldataT2: ndarray):
    """(3) Collects master element data from subelement level."""
    nE, _, _, nD = celldataT2.shape
    res = np.zeros((nE, 9, nD), dtype=celldataT2.dtype)
    for iE in prange(nE):
        for iSE in prange(4):
            for iNSE in prange(6):
                res[iE, topoT2[iSE, iNSE]] += celldataT2[iE, iSE, iNSE]
    return res


@njit(nogil=True, cache=__cache)
def nodal_data_T2(celldata: ndarray):
    """
    Handles loads defined on passive nodes in 3 steps:
    (1) Distribute master element data further to subelement level.
    (2) On every subelement, transforms passive loads to uniform
        body loads and integrates to active nodes.
    (3) Collects master element data from subelement level.
    """
    celldataT2 = distribute_nodal_data_T2(celldata)
    celldataT2 = transfer_passive_loads_T2(celldataT2)
    return collect_nodal_dataT2(celldataT2)


class Q5MV(Quadrilateral, Membrane, FiniteElement):

    compatible = False

    def stiffness_matrix(self, *args, topo=None, **kwargs):
        topo = self.nodes.to_numpy() if topo is None else topo
        ecoords = self.local_coordinates(topo=topo)
        C = self.model_stiffness_matrix()
        return stiffness_matrix_T2(C, ecoords)

    def global_dof_numbering(self, *args, topo=None, **kwargs):
        topo = self.nodes.to_numpy() if topo is None else topo
        return topo_to_gnum(topo[:, 4:], self.NDOFN)

    def distribute_nodal_data(self, data, key):
        super().distribute_nodal_data(data, key)
        if key == 'loads':
            celldata = self._wrapped['loads'].to_numpy()
            self._wrapped['loads'] = nodal_data_T2(celldata)
                
    def approximation_matrix(self, *args, **kwargs):
        return approximation_matrix_T2_bulk(self.ndf.to_numpy())
    
    def nodal_approximation_matrix(self, *args, **kwargs):
        return nodal_approximation_matrix_T2_bulk(self.ndf.to_numpy())
        



# -*- coding: utf-8 -*-
from neumann.numint import GaussPoints as Gauss
from polymesh.cells import L2 as Line

from .bernoulli import BernoulliBase as Bernoulli
from .gen.b2 import (shape_function_values_bulk, 
                     shape_function_derivatives_bulk)
from .elem import FiniteElement
from .metaelem import ABCFiniteElement as ABC

__all__ = ['Bernoulli2']


class Bernoulli2(ABC, Bernoulli, Line, FiniteElement):
    """
    Finite element class to handle 2-noded Bernoulli beams.
    """

    qrule = 'full'
    quadrature = {
        'full': Gauss(2),
        'selective': {
            (0, 1): 'full',
            (2): 'reduced'
        },
        'reduced': Gauss(1),
        'mass' : Gauss(4)
    }
    shpfnc = shape_function_values_bulk
    dshpfnc = shape_function_derivatives_bulk


# -*- coding: utf-8 -*-
import numpy as np
from numpy.linalg import inv
from numba import njit, prange

from ..utils import points_of_layers

__cache = True


@njit(nogil=True, parallel=True, cache=__cache)
def rotation_matrices(angles: np.ndarray):
    """
    Returns transformation matrices T_126 and T_45 for each angle.
    Angles are expected in radians.
    """
    nL = len(angles)
    T_126 = np.zeros((nL, 3, 3), dtype=angles.dtype)
    T_45 = np.zeros((nL, 2, 2), dtype=angles.dtype)
    for iL in prange(nL):
        a = angles[iL] * np.pi / 180
        T_126[iL, 0, 0] = np.cos(a)**2
        T_126[iL, 0, 1] = np.sin(a)**2
        T_126[iL, 0, 2] = -np.sin(2 * a)
        T_126[iL, 1, 0] = T_126[iL, 0, 1]
        T_126[iL, 1, 1] = T_126[iL, 0, 0]
        T_126[iL, 1, 2] = -T_126[iL, 0, 2]
        T_126[iL, 2, 0] = np.cos(a) * np.sin(a)
        T_126[iL, 2, 1] = -T_126[iL, 2, 0]
        T_126[iL, 2, 2] = np.cos(a)**2 - np.sin(a)**2
        T_45[iL, 0, 0] = np.cos(a)
        T_45[iL, 0, 1] = -np.sin(a)
        T_45[iL, 1, 1] = T_45[iL, 0, 0]
        T_45[iL, 1, 0] = -T_45[iL, 0, 1]
    return T_126, T_45


@njit(nogil=True, parallel=True, cache=__cache)
def material_stiffness_matrices(C_126: np.ndarray, C_45: np.ndarray, angles: np.ndarray):
    """
    Returns the components of the material stiffness matrices C_126 and C_45
    in the global system.
    """
    nL = len(C_126)
    C_126_g = np.zeros_like(C_126)
    C_45_g = np.zeros_like(C_45)
    R_126 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 2]], dtype=C_126_g.dtype)
    R_126_inv = np.linalg.inv(R_126)
    R_45 = np.array([[2, 0], [0, 2]], dtype=C_126_g.dtype)
    R_45_inv = np.linalg.inv(R_45)
    T_126, T_45 = rotation_matrices(angles)
    for iL in prange(nL):
        C_126_g[iL] = T_126[iL] @ C_126[iL] @ R_126_inv @ T_126[iL].T @ R_126
        C_45_g[iL] = T_45[iL] @ C_45[iL] @ R_45_inv @ T_45[iL].T @ R_45
    return C_126_g, C_45_g


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def shear_factors_MT(ABDS: np.ndarray, C_126: np.ndarray, z: np.ndarray):
    """
    # FIXME Should work with parallel = True, does not. Reason is
    propably a race condition due to using explicit parallel loops.
    """
    A11 = ABDS[0, 0]
    B11 = ABDS[0, 3]
    D11 = ABDS[3, 3]
    A22 = ABDS[1, 1]
    B22 = ABDS[1, 4]
    D22 = ABDS[4, 4]
    nL = C_126.shape[0]  # number of layers

    # calculate shear factors
    eta_x = 1 / (A11 * D11 - B11**2)
    eta_y = 1 / (A22 * D22 - B22**2)
    shear_factors = np.zeros((nL, 2, 3), dtype=ABDS.dtype)

    for iL in prange(nL-1):
        for iP in prange(2):
            zi0 = z[iL, iP]
            zi1 = z[iL, iP + 1]
            dsfx = - eta_x * C_126[iL, 0, 0] * \
                (0.5 * (zi1**2 - zi0**2) * A11 - (zi1 - zi0) * B11)
            shear_factors[iL, 0, iP + 1:] += dsfx
            shear_factors[iL + 1:, 0, :] += dsfx
            dsfy = - eta_y * C_126[iL, 1, 1] * \
                (0.5 * (zi1**2 - zi0**2) * A22 - (zi1 - zi0) * B22)
            # these slicings probably cause a race condition
            shear_factors[iL, 1, iP + 1:] += dsfy
            shear_factors[iL + 1:, 1, :] += dsfy
    # last layer
    iL = nL - 1
    for iP in prange(2):
        zi0 = z[iL, iP]
        zi1 = z[iL, iP + 1]
        dsfx = - eta_x * C_126[iL, 0, 0] * \
            (0.5 * (zi1**2 - zi0**2) * A11 - (zi1 - zi0) * B11)
        shear_factors[iL, 0, iP + 1:] += dsfx
        dsfy = - eta_y * C_126[iL, 1, 1] * \
            (0.5 * (zi1**2 - zi0**2) * A22 - (zi1 - zi0) * B22)
        shear_factors[iL, 1, iP + 1:] += dsfy
    shear_factors[iL, :, 2] = 0.
    return shear_factors


@njit(nogil=True, cache=__cache)
def shear_factors_ST(ABDS: np.ndarray, C_126: np.ndarray, z: np.ndarray):
    """
    Single-thread implementation of calculation of shear factors for
    multi-layer Mindlin shells.
    """
    A11 = ABDS[0, 0]
    B11 = ABDS[0, 3]
    D11 = ABDS[3, 3]
    A22 = ABDS[1, 1]
    B22 = ABDS[1, 4]
    D22 = ABDS[4, 4]
    nL = z.shape[0]  # number of layers

    # calculate shear factors
    eta_x = 1 / (A11 * D11 - B11**2)
    eta_y = 1 / (A22 * D22 - B22**2)
    shear_factors = np.zeros((nL, 2, 3), dtype=ABDS.dtype)

    for iL in range(nL):
        zi = z[iL]
        # first point through the thickness
        shear_factors[iL, 0, 0] = shear_factors[iL-1, 0, 2]
        shear_factors[iL, 1, 0] = shear_factors[iL-1, 1, 2]
        # second point through the thickness
        shear_factors[iL, 0, 1] = shear_factors[iL, 0, 0] - \
            eta_x * C_126[iL, 0, 0] * (0.5 * (zi[1]**2 - zi[0]**2) * A11 -
                                       (zi[1] - zi[0]) * B11)
        shear_factors[iL, 1, 1] = shear_factors[iL, 1, 0] - \
            eta_y * C_126[iL, 1, 1] * (0.5 * (zi[1]**2 - zi[0]**2) * A22 -
                                       (zi[1] - zi[0]) * B22)
        # third point through the thickness
        shear_factors[iL, 0, 2] = shear_factors[iL, 0, 0] - \
            eta_x * C_126[iL, 0, 0] * (0.5 * (zi[2]**2 - zi[0]**2) * A11 -
                                       (zi[2] - zi[0]) * B11)
        shear_factors[iL, 1, 2] = shear_factors[iL, 1, 0] - \
            eta_y * C_126[iL, 1, 1] * (0.5 * (zi[2]**2 - zi[0]**2) * A22 -
                                       (zi[2] - zi[0]) * B22)
    shear_factors[nL-1, :, 2] = 0.
    return shear_factors


@njit(nogil=True, parallel=True, cache=__cache)
def shear_correction_data(ABDS: np.ndarray, C_126: np.ndarray, C_45: np.ndarray,
                          bounds: np.ndarray):
    """
    FIXME : Results are OK, but a bit slower than expected when measured
    against the pure python implementation.
    """

    nL = bounds.shape[0]  # number of layers

    # z coordinate of 3 points per each layer
    z = points_of_layers(bounds, 3)

    # calculate shear factors
    shear_factors = shear_factors_ST(ABDS, C_126, z)

    # compile shear factors
    sf = np.zeros((nL, 2, 3), dtype=ABDS.dtype)
    for iL in prange(nL):
        monoms_inv = inv(np.array([[1, z, z**2]
                         for z in z[iL]], dtype=ABDS.dtype))
        sf[iL, 0] = monoms_inv @ shear_factors[iL, 0]
        sf[iL, 1] = monoms_inv @ shear_factors[iL, 1]

    # potential energy using constant stress distribution
    # and unit shear force
    pot_c_x = 0.5 / ABDS[6, 6]
    pot_c_y = 0.5 / ABDS[7, 7]

    # positions and weights of Gauss-points
    gP = np.array([-np.sqrt(3/5), 0, np.sqrt(3/5)], dtype=ABDS.dtype)
    gW = np.array([5/9, 8/9, 5/9], dtype=ABDS.dtype)

    # potential energy using parabolic stress distribution
    # and unit shear force
    pot_p_x, pot_p_y = 0., 0.
    for iL in prange(nL):
        dJ = 0.5 * (bounds[iL, 1] - bounds[iL, 0])
        Gxi = C_45[iL, 0, 0]
        Gyi = C_45[iL, 1, 1]
        for iG in prange(3):
            ziG = 0.5 * ((bounds[iL, 1] + bounds[iL, 0]) +
                         gP[iG] * (bounds[iL, 1] - bounds[iL, 0]))
            monoms = np.array([1, ziG, ziG**2], dtype=ABDS.dtype)
            sfx = np.dot(monoms, sf[iL, 0])
            sfy = np.dot(monoms, sf[iL, 1])
            pot_p_x += 0.5 * (sfx**2) * dJ * gW[iG] / Gxi
            pot_p_y += 0.5 * (sfy**2) * dJ * gW[iG] / Gyi
    kx = pot_c_x / pot_p_x
    ky = pot_c_y / pot_p_y

    return np.array([[kx, 0], [0, ky]]), shear_factors


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def stiffness_data(C_126: np.ndarray, C_45: np.ndarray, angles: np.ndarray,
                   bounds: np.ndarray):
    """
    FIXME Call
        Ks, sf = shear_correction_data(ABDS, C_126_g, C_45_g, bounds)
    is a bit slow for some reason.
    """
    ABDS = np.zeros((8, 8), dtype=C_126.dtype)
    nL = C_126.shape[0]
    C_126_g, C_45_g = material_stiffness_matrices(C_126, C_45, angles)
    for iL in prange(nL):
        ABDS[0:3, 0:3] += C_126_g[iL] * (bounds[iL, 1] - bounds[iL, 0])
        ABDS[0:3, 3:6] += (1 / 2) * C_126_g[iL] * \
            (bounds[iL, 1]**2 - bounds[iL, 0]**2)
        ABDS[3:6, 3:6] += (1 / 3) * C_126_g[iL] * \
            (bounds[iL, 1]**3 - bounds[iL, 0]**3)
        ABDS[6:8, 6:8] += C_45_g[iL] * (bounds[iL, 1] - bounds[iL, 0])
    ABDS[3:6, 0:3] = ABDS[0:3, 3:6]
    Ks, sf = shear_correction_data(ABDS, C_126_g, C_45_g, bounds)
    return ABDS, Ks, sf


@njit(nogil=True, cache=__cache)
def z_to_shear_factors(z, sfx, sfy):
    monoms = np.array([1, z, z**2], dtype=sfx.dtype)
    return np.dot(monoms, sfx), np.dot(monoms, sfy)



# -*- coding: utf-8 -*-
from typing import Callable
from numba import njit, prange
import numpy as np
from numpy import ndarray

from neumann.linalg import ReferenceFrame, inv

from polymesh.utils import cells_coords

__cache = True


@njit(nogil=True, cache=__cache)
def element_coordinates(coords: ndarray, topo: np.ndarray):
    return cells_coords(coords, topo)


def element_centers(ecoords: np.ndarray):
    return np.mean(ecoords, axis=1)


@njit(nogil=True, cache=__cache)
def loc_to_glob(pcoord: np.ndarray, gcoords: np.ndarray, shpfnc: Callable):
    """
    Local to global transformation for a single cell and point.

    Returns global coordinates of a point in an element, provided the global
    corodinates of the points of the element, an array of parametric
    coordinates and a function to evaluate the shape functions.

    Parameters
    ----------
    gcoords : (nNE, nD) ndarray
        2D array containing coordinates for every node of a single element.
            nNE : number of vertices of the element
            nD : number of dimensions of the model space

    pcoords : (nDP, ) ndarray
        1D array of parametric coordinates for a single point.
            nDP : number of dimensions of the parametric space

    shpfnc : Callable
        A function that evaluates shape function values at a point,
        specified with parametric coordinates.

    Returns
    -------
    (nD, ) ndarray
        Global cooridnates of the specified point.

    Notes
    -----
    'shpfnc' must be a numba-jitted function, that accepts a 1D array of
    exactly nDP number of components.
    
    """
    return gcoords.T @ shpfnc(pcoord)


@njit(nogil=True, parallel=True, cache=__cache)
def loc_to_glob_bulk(pcoords: np.ndarray, gcoords: np.ndarray,
                     shpfnc: Callable):
    """Local to global transformation for multiple cells and points.

    Returns global coordinates of points expressed in parametric coordinates.

    Parameters
    ----------
    gcoords : (nE, nNE, nD) ndarray
        3D array containing coordinates for every node of every element.
            nE : number of elements
            nNE : number of vertices of the element
            nD : number of dimensions of the model space

    pcoords : (nP, nDP) ndarray
        2D array of parametric coordinates for several points.
            nP : number of points
            nDP : number of dimensions of the parametric space

    shpfnc : Callable
        A function that evaluates shape function values at a point,
        specified with parametric coordinates.

    Returns
    -------
    (nE, nP, nD) ndarray
        Global cooridnates of the specified points in the specified elements.
        Dummy axes are collapsed, so the result may be 1, 2 or 3 dimensional,
        depending on the shapes of the input arrays.

    Notes
    -----
    'shpfnc' must be a numba-jitted function, that accepts a 1D array of
    exactly nDP number of components.
    """
    nE, nNE, nD = gcoords.shape
    nP, _ = pcoords.shape
    res = np.zeros((nE, nP, nD), dtype=gcoords.dtype)
    for iP in prange(nP):
        shp = shpfnc(pcoords[iP])
        for iE in prange(nE):
            for iD in prange(nD):
                for iNE in prange(nNE):
                    res[iE, iP, iD] += gcoords[iE, iNE, iD] * shp[iNE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def glob_to_loc(coord: np.ndarray, gcoords: np.ndarray, lcoords: np.ndarray,
                monomsfnc: Callable):
    """Global to local transformation for a single point and cell.

    Returns local coordinates of a point in an element, provided the global
    corodinates of the points of the element, an array of global
    coordinates and a function to evaluate the monomials of the shape
    functions.

    Parameters
    ----------
    gcoords : (nNE, nD) ndarray
        2D array containing coordinates for every node of a single element.
            nNE : number of vertices of the element
            nD : number of dimensions of the model space

    coord : (nD, ) ndarray
        1D array of global coordinates for a single point.
            nD : number of dimensions of the model space

    lcoords : (nNE, nDP) ndarray
        2D array of local coordinates of the parametric element.
            nNE : number of vertices of the element
            nDP : number of dimensions of the parametric space

    monomsfnc : Callable
        A function that evaluates monomials of the shape functions at a point
        specified with parametric coordinates.

    Returns
    -------
    (nDP, ) ndarray
        Parametric cooridnates of the specified point.

    Notes
    -----
    'shpfnc' must be a numba-jitted function, that accepts a 1D array of
    exactly nDP number of components.
    """
    nNE = gcoords.shape[0]
    monoms = np.zeros((nNE, nNE), dtype=coord.dtype)
    for i in prange(nNE):
        monoms[:, i] = monomsfnc(gcoords[i])
    coeffs = inv(monoms)
    shp = coeffs @ monomsfnc(coord)
    return lcoords.T @ shp


@njit(nogil=True, parallel=True, cache=__cache)
def glob_to_loc_bulk(coords: np.ndarray, gcoords: np.ndarray,
                     lcoords: np.ndarray, monomsfnc: Callable):
    """Global to local transformation for multiple cells and points.

    Returns local coordinates of a point in an element, provided the global
    corodinates of the points of the element, an array of global
    coordinates and a function to evaluate the monomials of the shape
    functions.

    Parameters
    ----------
    gcoords : (nE, nNE, nD) ndarray
        3D array containing coordinates for every node of several elements.
            nE : number of elements
            nNE : number of vertices of the element
            nD : number of dimensions of the model space

    coords : (nP, nD) ndarray
        2D array of global coordinates for several points.
            nP : number of points
            nD : number of dimensions of the model space

    lcoords : (nNE, nDP) ndarray
        2D array of local coordinates of the parametric element.
            nNE : number of vertices of the element
            nDP : number of dimensions of the parametric space

    monomsfnc : Callable
        A function that evaluates monomials of the shape functions at a point
        specified with parametric coordinates.

    Returns
    -------
    (nP, nE, nDP) ndarray
        Parametric cooridnates of the specified points in all elements.

    Notes
    -----
    'shpfnc' must be a numba-jitted function, that accepts a 1D array of
    exactly nDP number of components, where nDP is the number
    of paramatric cooridnate dimensions.
    """
    nE, nNE, _ = gcoords.shape
    nP = len(coords)
    nDP = lcoords.shape[1]
    lcoords = lcoords.T
    res = np.zeros((nP, nE, nDP), dtype=gcoords.dtype)
    monoms = np.zeros((nNE, nNE), dtype=gcoords.dtype)
    shp = np.zeros(nNE, dtype=gcoords.dtype)
    coeffs = np.zeros_like(monoms)
    for iE in prange(nE):
        for iNE in prange(nNE):
            monoms[:, iNE] = monomsfnc(gcoords[iE, iNE])
        coeffs[:, :] = inv(monoms)
        for iP in prange(nP):
            shp[:] = coeffs @ monomsfnc(coords[iP])
            res[iP, iE, :] = lcoords @ shp
    return res


@njit(nogil=True, cache=__cache)
def pip(coord: np.ndarray, gcoords: np.ndarray, lcoords: np.ndarray,
        monomsfnc: Callable, shpfnc: Callable, tol=1e-12):
    """Point-in-polygon test for a single cell and point.

    Performs the point-in-poligon test for a single point and cell.
    False means that the point is outside of the cell, True means the point
    is either inside, or on the boundary of the cell.

    Parameters
    ----------
    gcoords : (nNE, nD) ndarray
        2D array containing coordinates for every node of a single element.
            nNE : number of vertices of the element
            nD : number of dimensions of the model space

    coord : (nD, ) ndarray
        1D array of global coordinates for a single point.
            nD : number of dimensions of the model space

    lcoords : (nNE, nDP) ndarray
        2D array of local coordinates of the parametric element.
            nNE : number of vertices of the element
            nDP : number of dimensions of the parametric space

    monomsfnc : Callable
        A function that evaluates monomials of the shape functions at a point
        specified with parametric coordinates.

    shpfnc : Callable
        A function that evaluates shape function values at a point,
        specified with parametric coordinates.

    Returns
    -------
    bool
        False if points is outside of the cell, True otherwise.

    Notes
    -----
    'shpfnc'  and 'monomsfnc' must be numba-jitted functions, that accept
    a 1D array of exactly nDP number of components, where nDP is the number
    of paramatric cooridnate dimensions.
    """
    limit = 1 + tol
    loc = glob_to_loc(coord, gcoords, lcoords, monomsfnc)
    return np.all(shpfnc(loc) <= limit)


@njit(nogil=True, parallel=True, cache=__cache)
def int_domain(points: np.ndarray, qpos: np.ndarray, qweight: np.ndarray,
               dshpfnc: Callable, weight: float = 1.0):
    """
    Returns the measure (length, area or volume in 1d, 2d and 3d) of a 
    single domain.
    """
    volume = 0.
    points = np.transpose(points)
    nG = len(qweight)
    for iG in prange(nG):
        jac = points @ dshpfnc(qpos[iG])
        djac = np.linalg.det(jac)
        volume += qweight[iG] * djac
    return volume * weight


@njit(nogil=True, parallel=True, cache=__cache)
def int_domains(ecoords: np.ndarray, qpos: np.ndarray,
                qweight: np.ndarray, dshpfnc: Callable):
    """
    Returns the measure (length, area or volume in 1d, 2d and 3d) of 
    several domains.
    """
    nE = ecoords.shape[0]
    res = np.zeros(nE, dtype=ecoords.dtype)
    nG = len(qweight)
    for iG in prange(nG):
        dshp = dshpfnc(qpos[iG])
        for i in prange(nE):
            jac = ecoords[i].T @ dshp
            djac = np.linalg.det(jac)
            res[i] += qweight[iG] * djac
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def stiffness_matrix(C: np.ndarray, ecoords: np.ndarray, qpos: np.ndarray,
                     qweight: np.ndarray, dshpfnc: Callable,
                     sdmfnc: Callable, nDOFN=3):
    nTOTV = len(ecoords) * nDOFN
    res = np.zeros((nTOTV, nTOTV), dtype=C.dtype)
    points = np.transpose(ecoords)
    nG = len(qweight)
    for iG in prange(nG):
        dshp = dshpfnc(qpos[iG])
        jac = points @ dshp
        djac = np.linalg.det(jac) * qweight[iG]
        B = sdmfnc(jac, dshp)
        res += B.T @ C @ B * djac
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def stiffness_matrix_bulk(C: np.ndarray, ecoords: np.ndarray,
                          qpos: np.ndarray, qweight: np.ndarray,
                          dshpfnc: Callable, sdmfnc: Callable, nDOFN=3):
    nE, nNODE, _ = ecoords.shape
    nTOTV = nNODE * nDOFN
    res = np.zeros((nE, nTOTV, nTOTV), dtype=C.dtype)
    nG = len(qweight)
    for iG in prange(nG):
        dshp = dshpfnc(qpos[iG])
        for iE in prange(nE):
            jac = np.transpose(ecoords[iE]) @ dshp
            djac = np.linalg.det(jac) * qweight[iG]
            B = sdmfnc(jac, dshp)
            res[iE] += B.T @ C[iE] @ B * djac
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def strain_displacement_matrix_bulk2(B: ndarray, djac: ndarray, w: ndarray):
    nE, nG, nSTRE, nTOTV = B.shape
    res = np.zeros((nE, nSTRE, nTOTV), dtype=B.dtype)
    for iG in range(nG):
        for iE in prange(nE):
            res[iE] += B[iE, iG] * djac[iE, iG] * w[iG]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def unit_strain_load_vector_bulk(D: ndarray, B: ndarray):
    nE, nSTRE, nTOTV = B.shape
    res = np.zeros((nE, nTOTV, nSTRE), dtype=D.dtype)
    for iE in prange(nE):
        res[iE] = B[iE].T @ D[iE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def strain_load_vector_bulk(BTD: ndarray, strains: ndarray):
    nE, nTOTV, nSTRE = BTD.shape
    nE, nRHS, nSTRE = strains.shape
    res = np.zeros((nE, nTOTV, nRHS), dtype=BTD.dtype)
    for iE in prange(nE):
        for iRHS in prange(nRHS):
            res[iE, :, iRHS] = BTD[iE] @ strains[iE, iRHS, :]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def stiffness_matrix_bulk2(D: ndarray, B: ndarray, djac: ndarray, w: ndarray):
    nE, nG = djac.shape
    nTOTV = B.shape[-1]
    res = np.zeros((nE, nTOTV, nTOTV), dtype=D.dtype)
    for iG in range(nG):
        for iE in prange(nE):
            res[iE] += B[iE, iG].T @ D[iE] @ B[iE, iG] * djac[iE, iG] * w[iG]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def mass_matrix_bulk(N: ndarray, dens: ndarray, weights: ndarray,
                     djac: ndarray, w: ndarray):
    """
    Evaluates the mass matrix of several elements from evaluations at the
    Gauss points of the elements. The densities are assumed to be constant
    over the elements, and be represented by the 1d float numpy array 'dens'.

    Parameters
    ----------
    N : 4d numpy float array
        Evaluations of the shape function matrix according to the type of 
        the element for every cell and Gauss point.

    dens : 1d numpy float array
        1d float array of densities of several elements.

    weights : 1d numpy float array
        Array of weighting factors for the cells (eg. areas, volumes) with 
        the same shape as 'dens'. 

    w : 1d numpy float array
        Weights of the Gauss points

    djac : 2d numpy float array
        Jacobi determinants for every element and Gauss point.

    nE : number of elements
    nG : number of Gauss points
    nNE : number of nodes per element
    nDOF : number of dofs per node

    """
    nE, nG = djac.shape
    nTOTV = N.shape[-1]
    res = np.zeros((nE, nTOTV, nTOTV), dtype=N.dtype)
    for iG in range(nG):
        for iE in prange(nE):
            res[iE] += dens[iE] * weights[iE] * \
                N[iE, iG].T @ N[iE, iG] * djac[iE, iG] * w[iG]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def compliance_bulk(K_bulk: np.ndarray, dofsol1d: np.ndarray,
                    gnum: np.ndarray):
    """
    Returns the total compliance of the structure, that is
        u.T @ K @ u
    """
    nE, nTOTV, _ = K_bulk.shape
    res = np.zeros(nE, dtype=K_bulk.dtype)
    for iE in prange(nE):
        for i in prange(nTOTV):
            for j in prange(nTOTV):
                res[iE] += K_bulk[iE, i, j] * \
                    dofsol1d[gnum[iE, i]] * dofsol1d[gnum[iE, j]]
    return res


@njit(nogil=True, cache=__cache)
def elastic_strain_energy_bulk(K_bulk: np.ndarray, dofsol1d: np.ndarray,
                               gnum: np.ndarray):
    """
    Returns the total elastic strain energy of the structure, that is 
        0.5 * u.T @ K @ u
    """
    return compliance_bulk(K_bulk, dofsol1d, gnum) / 2


@njit(nogil=True, parallel=True, cache=__cache)
def approx_dof_solution(dofsol1d: np.ndarray, gnum: np.ndarray,
                        shpmat: np.ndarray):
    nDOFN, nTOTV = shpmat.shape
    res = np.zeros((nDOFN), dtype=dofsol1d.dtype)
    for j in prange(nDOFN):
        for k in prange(nTOTV):
            res[j] += shpmat[j, k] * dofsol1d[gnum[k]]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def approx_dof_solution_bulk(dofsol1d: np.ndarray, gnum: np.ndarray,
                             shpmat: np.ndarray):
    nE = len(gnum)
    nDOFN, nTOTV = shpmat.shape
    res = np.zeros((nE, nDOFN), dtype=dofsol1d.dtype)
    for i in prange(nE):
        for j in prange(nDOFN):
            for k in prange(nTOTV):
                res[i, j] += shpmat[j, k] * dofsol1d[gnum[i, k]]
    return res


@njit(nogil=True, cache=__cache)
def element_dof_solution(dofsol1d: np.ndarray, gnum: np.ndarray):
    return dofsol1d[gnum]


@njit(nogil=True, parallel=True, cache=__cache)
def element_dof_solution_bulk(dofsol1d: ndarray, gnum: ndarray):
    nE, nEVAB = gnum.shape
    res = np.zeros((nE, nEVAB), dtype=dofsol1d.dtype)
    for i in prange(nE):
        res[i, :] = dofsol1d[gnum[i, :]]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def element_dof_solution_bulk_multi(dofsol: ndarray, gnum: ndarray):
    nRHS = dofsol.shape[-1]
    nE, nEVAB = gnum.shape
    res = np.zeros((nRHS, nE, nEVAB), dtype=dofsol.dtype)
    for i in prange(nE):
        for j in prange(nRHS):
            res[j, i, :] = dofsol[gnum[i, :], j]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def avg_nodal_data_bulk(data: np.ndarray, topo: np.ndarray):
    nE, nNE, nD = data.shape
    nP = np.max(topo) + 1
    res = np.zeros((nP, nD), dtype=data.dtype)
    count = np.zeros((nP), dtype=topo.dtype)
    for iE in prange(nE):
        for jNE in prange(nNE):
            res[topo[iE, jNE]] += data[iE, jNE]
            count[topo[iE, jNE]] += 1
    for iP in prange(nP):
        res[iP] = res[iP] / count[iP]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def stresses_at_point_bulk(C: np.ndarray, dofsol1d: np.ndarray,
                           ecoords: np.ndarray, gnum: np.ndarray,
                           dshp: np.ndarray, sdmfnc: Callable):
    nE = len(gnum)  # number of elements
    nSTRE = C.shape[-1]  # number of generalized forces
    esol = element_dof_solution_bulk(dofsol1d, gnum)
    res = np.zeros((nE, nSTRE), dtype=C.dtype)
    for i in prange(nE):
        jac = ecoords[i].T @ dshp
        B = sdmfnc(jac, dshp)
        res[i] = C[i] @ B @ esol[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def stresses_at_cells_nodes(C: np.ndarray, dofsol1d: np.ndarray,
                            ecoords: np.ndarray, gnum: np.ndarray,
                            dshp: np.ndarray, sdmfnc: Callable):
    nE, nNODE, _ = ecoords.shape  # numbers of elements and element nodes
    nSTRE = C.shape[-1]  # number of generalized forces
    res = np.zeros((nE, nNODE, nSTRE), dtype=C.dtype)
    for i in prange(nNODE):
        res[:, i, :] = stresses_at_point_bulk(C, dofsol1d, ecoords, gnum,
                                              dshp[i], sdmfnc)
    return res


# -*- coding: utf-8 -*-
import numpy as np

from polymesh.abc import ABCMeta_MeshData
from polymesh.cell import PolyCell

from ..dofmap import DOF


class FemMixin:
    
    dofs = ()
    dofmap = ()
            
    # optional
    qrule: str = None
    quadrature = None

    # advanced settings
    compatible = True
    
    # must be reimplemented
    NNODE: int = None  # number of nodes, normally inherited

    # from the mesh object
    # it affects sizing of all kinds of intermediate variables
    # during calculations (eg. element transformation matrix)
    NDOFN: int = None  # numper of dofs per node, normally

    # inherited form the model object
    NDIM: int = None  # number of geometrical dimensions,
    # normally inherited from the mesh object

    # inherited form the model object
    NSTRE: int = None  # number of internal force components,
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    @property    
    def geometry(self):
        pd, cd = self.pdb, self.db 
        return self.__class__.Geometry(pointdata=pd, celldata=cd)
    
    @property
    def pdb(self):
        return self.pointdata
    
    @property
    def db(self):
        return self._wrapped
    
    def topology(self):
        raise NotImplementedError
    
    # !TODO : this should be implemented at geometry
    @classmethod
    def lcoords(cls, *args, **kwargs):
        raise NotImplementedError

    # !TODO : this should be implemented at geometry
    @classmethod
    def lcenter(cls, *args, **kwargs):
        raise NotImplementedError
    
    def local_coordinates(self, *args, frames=None, _topo=None, **kwargs):
        # implemented at PolyCell
        raise NotImplementedError

    def points_of_cells(self):
        # implemented at PolyCell
        raise NotImplementedError

    # !TODO : this should be implemented at geometry
    # !TODO : can be reimplemented if the element is not IsoP
    def jacobian_matrix(self, *args, **kwargs):
        raise NotImplementedError

    # !TODO : this should be implemented at geometry
    # !TODO : can be reimplemented if the element is not IsoP
    def jacobian(self, *args, **kwargs):
        raise NotImplementedError

    # !TODO : this should be implemented at model
    def strain_displacement_matrix(self, *args, **kwargs):
        raise NotImplementedError

    # !TODO : this should be implemented at geometry
    def shape_function_values(self, *args, **kwargs):
        raise NotImplementedError

    # !TODO : this should be implemented at geometry
    # !TODO : can be reimplemented if the element is not IsoP
    def shape_function_matrix(self, *args, **kwargs):
        raise NotImplementedError

    # !TODO : this should be implemented at geometry
    def shape_function_derivatives(self, *args, **kwargs):
        raise NotImplementedError

    # !TODO : this should be implemented at model
    def model_stiffness_matrix(self, *args, **kwargs):
        raise NotImplementedError

    # !TODO : this should be implemented at model
    def stresses_at_centers(self, *args, **kwargs):
        raise NotImplementedError

    # !TODO : this should be implemented at model
    def stresses_at_cells_nodes(self, *args, **kwargs):
        raise NotImplementedError

    def direction_cosine_matrix(self):
        return None
    
    def integrate_body_loads(self, *args, **kwargs):
        raise NotImplementedError
    
    def masses(self, *args, **kwargs):
        raise NotImplementedError
    
    def volumes(self, *args, **kwargs):
        raise NotImplementedError
    
    def densities(self, *args, **kwargs):
        raise NotImplementedError
        
    def weights(self, *args, **kwargs):
        raise NotImplementedError
    
    def _postproc_local_internal_forces_(self, forces, *args, points, rng, cells, **kwargs):
        """
        The aim of this function os to guarantee a standard shape of output, that contains
        values for each of the 6 internal force compoents, irrespective of the kinematical
        model being used.

        Example use case : the Bernoulli beam element, where, as a consequence of the absence
        of shear forces, there are only 4 internal force components, and the shear forces
        must be calculated a-posteriori.

        Parameters
        ----------
        forces : numpy array of shape (nE, nP, nSTRE, nRHS)
            4d float array of internal forces for many elements, evaluation points,
            and load cases. The number of force components (nSTRE) is dictated by the
            reduced kinematical model (eg. 4 for the Bernoulli beam).

        dofsol (nE, nEVAB, nRHS)

        Notes
        -----
        Arguments are based on the reimplementation in the Bernoulli base element.
        
        Returns
        -------
        numpy.ndarray 
            shape : (nE, nP, 6, nRHS)
        
        """
        return forces
    


class FemModel(FemMixin):
    ...
        

class FemModel1d(FemModel):
    ...


class MetaFiniteElement(ABCMeta_MeshData):
    """
    Python metaclass for safe inheritance. Throws a TypeError
    if a method tries to shadow a definition in any of the base
    classes.
    """

    def __init__(self, name, bases, namespace, *args, **kwargs):
        super().__init__(name, bases, namespace, *args, **kwargs)

    def __new__(metaclass, name, bases, namespace, *args, **kwargs):     
        cls = super().__new__(metaclass, name, bases, namespace, *args, **kwargs)    
        for base in bases:                
            if issubclass(base, PolyCell):
                cls.Geometry = base  
            if issubclass(base, FemModel):
                if len(base.dofs) > 0:
                    cls.Model = base
                    cls.dofs = base.dofs
                    cls.NDOFN = len(cls.dofs)
                    cls.dofmap = np.array(DOF.dofmap(cls.dofs), dtype=int)
        return cls
    

class ABCFiniteElement(metaclass=MetaFiniteElement):
    """
    Helper class that provides a standard way to create an ABC using
    inheritance.
    """
    __slots__ = ()
    dofs = ()
    dofmap = ()  



# -*- coding: utf-8 -*-
import numpy as np

from polymesh.cells import T3 as Triangle

from ..model.membrane import Membrane
from ..model.plate import Plate

from .elem import FiniteElement
from .metaelem import ABCFiniteElement as ABC



class CSTM(ABC, Membrane, Triangle, FiniteElement):
    """
    The constant-strain triangle (a.k.a., CST triangle, Turner triangle)
    """

    qrule = 'full'
    quadrature = {
        'full': (np.array([[1/3, 1/3]]), np.array([1/2])),
    }


class CSTP(ABC, Plate, Triangle, FiniteElement):
    """
    The constant-strain triangle (a.k.a., CST triangle, Turner triangle)
    """

    qrule = 'full'
    quadrature = {
        'full': (np.array([[1/3, 1/3]]), np.array([1/2])),
    }




# -*- coding: utf-8 -*-
from neumann.numint import GaussPoints as Gauss

from polymesh.cells import H27 as HexaHedron

from ..model.solid3d import Solid3d

from .elem import FiniteElement
from .metaelem import ABCFiniteElement as ABC


class H27(ABC, Solid3d, HexaHedron, FiniteElement):

    qrule = 'selective'
    quadrature = {
        'full': Gauss(3, 3, 3),
        'selective': {
            (0, 1, 2): 'full',
            (3, 4, 5): 'reduced'
        },
        'reduced': Gauss(2, 2, 2)
    }



# -*- coding: utf-8 -*-
from typing import Union, Tuple

from scipy.sparse import coo_matrix
from scipy.sparse import spmatrix
import numpy as np
from numpy import ndarray
from numba import njit, prange

from neumann import squeeze
from neumann.logical import isintegerarray, isfloatarray, isboolarray
from neumann.array import bool_to_float, atleastnd, matrixform
from neumann.linalg.sparse.utils import lower_spdata, upper_spdata

from .utils import (nodes2d_to_dofs1d, irows_icols_bulk,
                    nodal_mass_matrix_data, min_stiffness_diagonal)
from .imap import index_mappers, box_spmatrix, box_rhs, box_dof_numbering
from .constants import DEFAULT_DIRICHLET_PENALTY

__cache = True


def fem_coeff_matrix_coo(A: ndarray, *args, inds: ndarray = None,
                         rows: ndarray = None, cols: ndarray = None,
                         N: int = None):
    """
    Returns the coefficient matrix in sparse 'coo' format.
    Index mapping is either specified with `inds` or with
    both `rows` and `cols`.

    If `lower` or `upper` is provided as a positional argument, 
    the result is returned as a lower- or upper-triangular matrix, 
    respectively. 

    Parameters
    ----------
    A : np.ndarray
        Element coefficient matrices as a 3d array. The lengths if the last
        two axes must be the same.

    inds : np.ndarray, Optional
        Global indices of all quantities of an element as a 2d integer array. 
        Length of the second axis must match the lengths of the 2nd and 3rd
        axes of 'A'.
        Default is None.

    rows : np.ndarray, Optional
        Global numbering of the rows as an 1d integer array. Length must
        equal the flattened length of 'A'. Default is None.

    cols : np.ndarray, Optional
        Global numbering of the columns as an 1d integer array. Length must
        equal the flattened length of 'A'. Default is None.

    N : int, Optional
        Total number of coefficients in the system. If not provided, 
        it is inferred from index data.

    Returns
    -------
    scipy.sparse.coo_matrix
        Coefficient matrix in sparse coo format (see scipy doc for the details).

    """
    if N is None:
        assert inds is not None, "shape or `inds` must be provided!."
        N = inds.max() + 1
    if rows is None:
        assert inds is not None, "Row and column indices (`rows` and `cols`) " \
            "or global index numbering array (inds) must be provided!"
        rows, cols = irows_icols_bulk(inds)

    data, rows, cols = A.flatten(), rows.flatten(), cols.flatten()

    if len(args) > 0:
        if 'lower' in args:
            data, rows, cols = lower_spdata(data, rows, cols)
        elif 'upper' in args:
            data, rows, cols = upper_spdata(data, rows, cols)

    return coo_matrix((data, (rows, cols)), shape=(N, N))


def _build_nodal_data(values: ndarray, *, inds: ndarray = None, N: int = None):
    """
    Returns nodal data of any sort in standard form.

    The data is given with 'values' and applies to the node indices specified
    by 'inds'. At all times, the function must be informed about the total number of 
    nodes in the model, to return an output with a correct shape. If 'inds' is None, it 
    is assumed that 'values' contains data for each node in the model, and therefore
    the length of 'values' is the number of nodes in the model. Otherwise, the number of 
    nodes is inferred from 'inds' and 'N' (see the doc). 

    Parameters
    ----------
    values : numpy.ndarray
        2d or 3d numpy array of floats, that specify some sort of data 
        imposed on nodes for one or several datasets. If it is a 3d array, the length
        of the third axis equals the number of datasets.

    inds : numpy.ndarray, Optional
        1d numpy integer array specifying global indices for the rows of 'values'. 

    N : int, Optional
        The overall number of nodes in the model. If not provided, we assume that
        it equals the highest index in 'inds' + 1 (this assumes zero-based indexing).

    Notes
    -----
    The function returns the values as a 2d array, even if the input array was only 
    2 dimensional, which suggests a single load case. 

    Returns
    -------
    indices : numpy.ndarray
        1d numpy integer array of npde indices

    nodal_loads : numpy.ndarray 
        2d numpy float array of shape (nN * nDOF, nX), where nN, nDOF and nX
        are the number of nodes, degrees of freedom and datasets.

    N : int
        The overall size of the equation system, as derived from the input.

    """
    assert isfloatarray(values)

    # node based definition with multiple RHS (at least 1)
    values = atleastnd(values, 3, back=True)  # (nP, nDOF, nRHS)
    if inds is None:
        inds = np.arange(len(values))  # node indices
    else:
        assert isintegerarray(inds)
        assert len(values) == len(inds)

    # transform to nodal defintion
    inds, values = nodes2d_to_dofs1d(inds, values)

    N = inds.max() + 1 if N is None else N
    return inds, values, N


@squeeze(True)
def nodal_load_vector(values: ndarray, **kwargs) -> ndarray:
    """
    Assembles the right-hand-side of the global equation system.

    Parameters
    ----------
    values : numpy.ndarray, Optional
        2d or 3d numpy array of floats. If 3d, it is assumed that 
        the last axis runs along load cases.

    Returns
    -------
    numpy.ndarray
        The load vector as a 1d or 2d numpy array of floats.

    """
    inds, values, N = _build_nodal_data(values, **kwargs)
    nRHS = values.shape[-1]
    f = np.zeros((N, nRHS))
    f[inds] = values
    return f


@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def assemble_load_vector(values: ndarray, gnum: ndarray, N: int = -1):
    """
    Returns global dof numbering based on element 
    topology data.

    Parameters
    ----------
    values : numpy.ndarray 
        3d numpy float array of shape (nE, nEVAB, nRHS), representing 
        element data. The length of the second axis matches the the number of
        degrees of freedom per cell.

    gnum : int
        Global indices of local degrees of freedoms of elements.

    N : int, Optional
        The number of total unknowns in the system. Must be specified correcly,
        to get a vector the same size of the global system. If not specified, it is
        inherited from 'gnum' (as 'gnum.max() + 1'), but this can lead to a chopped 
        array.

    Returns
    -------
    numpy.ndarray
        2d numpy array of integers with a shape of (N, nRHS), where nRHS is the number
        if right hand sizes (load cases).

    """
    nE, nEVAB, nRHS = values.shape
    if N < 0:
        N = gnum.max() + 1
    res = np.zeros((N, nRHS), dtype=values.dtype)
    for i in range(nE):
        for j in range(nEVAB):
            for k in prange(nRHS):
                res[gnum[i, j], k] += values[i, j, k]
    return res


def essential_penalty_factor_matrix(fixity: ndarray = None, *, inds: ndarray = None,
                                    N: int = None, eliminate_zeros: bool = True,
                                    sum_duplicates: bool = True) -> coo_matrix:
    """
    Returns the penalty factor matrix. This matrix is the basis of several penalty
    matrices used to enforce constraints on the primary variables. 

    At all times, the function must be informed about the total number of nodes in the 
    model, to return an output with a correct shape. If 'inds' is None, it is assumed that 
    'fixity' contains data for each node in the model, and therefore the length of 'fixity' 
    is the number of nodes in the model. You can also use 'N' to indicate the number of 
    nodes, in which case entries according to 'inds' are marked as constrained.

    Parameters
    ----------
    fixity : numpy.ndarray, Optional
        Fixity information as a 2d boolean array.

    inds : numpy.ndarray, Optional
        1d numpy integer array specifying node indices. 

    N : int, Optional
        The overall number of nodes in the model. If not provided, we assume that
        it equals the highest index in 'inds' + 1 (this assumes zero-based indexing).

    """
    if fixity is not None:
        assert isboolarray(fixity)
    else:
        if isintegerarray(inds):
            if not isinstance(N, int):
                N = inds.max() + 1  # assumes zero based indexing
            fixity = np.zeros(N, dtype=bool)
            fixity[inds] = True
    assert isboolarray(fixity), "Invalid input!"
    fixity = bool_to_float(fixity, 1.0, 0.0)
    fixity = atleastnd(fixity, 3, back=True)
    inds, fixity, N = _build_nodal_data(fixity, inds=inds, N=N)
    K = coo_matrix((fixity[:, 0], (inds, inds)), shape=(N, N))
    if eliminate_zeros:
        K.eliminate_zeros()
    if sum_duplicates:
        K.sum_duplicates()
    return K


def essential_penalty_matrix(fixity: ndarray = None, *, inds: ndarray = None,
                             pfix: float = DEFAULT_DIRICHLET_PENALTY, 
                             eliminate_zeros: bool = True,
                             sum_duplicates: bool = True) -> coo_matrix:
    """
    Returns the sparse, COO format penalty matrix, equivalent of 
    a Courant-type penalization of the essential boundary conditions.

    Parameters
    ----------
    fixity : numpy.ndarray, Optional
        Fixity information as a 2d float or boolean array.

    inds : numpy.ndarray, Optional
        1d integer array specifying global indices for the rows of 'values'. 

    pfix : float, Optional
        Penalty value for fixed dofs. It is used to transform boolean penalty
        data, or to make up for missing values (e.g. only indices are provided).
        Default is `sigmaepsilon.solid.fem.constants.DEFAULT_DIRICHLET_PENALTY`.

    eliminate_zeros : bool, Optional
        Eliminates zeros from the matrix. Default is True.

    eliminate_zeros : bool, Optional
        Summs duplicate entries in the matrix. Default is True.

    Returns
    -------
    scipy.sparse.coo_matrix
        The penalty matrix in sparse COO format.

    """
    if fixity is not None:
        assert isinstance(fixity, np.ndarray)
    else:
        if isintegerarray(inds):
            if not isinstance(N, int):
                N = inds.max() + 1  # assumes zero based indexing
            fixity = np.zeros(N, dtype=bool)
            fixity[inds] = True
    if isboolarray(fixity):
        fixity = bool_to_float(fixity, pfix)
    assert isfloatarray(fixity)
    fixity = atleastnd(fixity, 3, back=True)  # (nP, nDOF, nX)
    inds, fixity, N = _build_nodal_data(fixity, inds=inds)
    K = coo_matrix((fixity[:, 0], (inds, inds)), shape=(N, N))
    if eliminate_zeros:
        K.eliminate_zeros()
    if sum_duplicates:
        K.sum_duplicates()
    return K


def nodal_mass_matrix(*args, values: ndarray = None, eliminate_zeros: bool = True,
                      sum_duplicates: bool = True, ndof: int = 6, **kwargs) -> coo_matrix:
    """
    Returns the diagonal mass matrix resulting form nodal masses in scipy 
    COO format.

    Parameters
    ----------
    See the documentation of 'build_fem_ebc_data' for the discription
    of the possible arguments.

    Returns
    -------
    scipy.sparse.coo_matrix
        The mass matrix in sparse COO format.

    """
    values = nodal_mass_matrix_data(values, ndof)
    inds, vals, N = _build_nodal_data(values, **kwargs)
    M = coo_matrix((vals[:, 0], (inds, inds)), shape=(N, N))
    if eliminate_zeros:
        M.eliminate_zeros()
    if sum_duplicates:
        M.sum_duplicates()
    return M


def estimate_stiffness_penalty(K:Union[ndarray, spmatrix], fixity:ndarray, N:int):
    """
    Returns a sugegsted value for the stiffness penalty, to account
    for essential boundary conditions.
    
    :math:`\\mathbf{M}`
    
    .. math::
        :nowrap:

        \\begin{equation}
            \\mathbf{A} \\mathbf{x} = \\lambda \\mathbf{x},
        \\end{equation}
    
    Parameters
    ----------
    K : Union[ndarray, spmatrix]
        The stiffness matrix as a dense or a sparse array.
    
    N : int
        The number of degrees of freedom in the model.
        
    References
    ----------
    .. [1] B. Nour-Omid, P. Wriggers "A two-level iteration method for 
       solution of contact problems." Computer Methods in Applied Mechanics
       and Engineering, vol. 54, pp. 131-144, 1986.
        
    """
    eps = np.finfo(float).eps
    kmin = min_stiffness_diagonal(K)
    return kmin / np.sqrt(N * eps)


@njit(nogil=True, parallel=True, cache=__cache)
def _pull_submatrix(A: ndarray, r: ndarray, c: ndarray) -> ndarray:
    nR = r.shape[0]
    nC = c.shape[0]
    res = np.zeros((nR, nC), dtype=A.dtype)
    for i in prange(nR):
        for j in prange(nC):
            res[i, j] = A[r[i], c[j]]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _pull_subvector(v: ndarray, i: ndarray) -> ndarray:
    nI = i.shape[0]
    res = np.zeros((nI, v.shape[-1]), dtype=v.dtype)
    for j in prange(nI):
        res[j, :] = v[i[j], :]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _push_submatrix(A: ndarray, Asub: ndarray, r: ndarray, c: ndarray):
    nR = r.shape[0]
    nC = c.shape[0]
    for i in prange(nR):
        for j in prange(nC):
            A[r[i], c[j]] = Asub[i, j]
            

@njit(nogil=True, parallel=True, cache=__cache)
def _push_subvector(v: ndarray, vsub: ndarray, i: ndarray):
    nI = i.shape[0]
    for j in prange(nI):
        v[i[j], :] = vsub[j, :]


@njit(nogil=True, parallel=True, cache=__cache)
def condensate_Kf_bulk(K: ndarray, f: ndarray, 
                       fixity: ndarray) -> Tuple[ndarray, ndarray]:
    """
    Returns the condensed coefficient matrices representing constraints
    on the internal forces of the elements (eg. hinges).

    Currently this solution is only able to handle two states, being totally free 
    and being fully constrained. The fixity values are expected to be numbers between
    0 and 1, where dofs with a factor > 0.5 are assumed to be the constrained ones.

    Parameters
    ----------
    K : numpy.ndarray
        3d float array, the stiffness matrices of several elements of the same kind.

    factors: numpy.ndarray
        2d boolean array of connectivity factors for each dof of several elements.

    Returns
    -------
    numpy.ndarray
        The condensed stiffness matrices with the same shape as `K`.
        
    numpy.ndarray
        The condensed load vectors with the same shape as `f`.

    """
    nE = K.shape[0]
    K_out = np.zeros_like(K)
    f_out = np.zeros_like(f)
    for iE in prange(nE):
        b = np.where(fixity[iE])[0]
        i = np.where(~fixity[iE])[0]
        # stiffness matrix
        Kbb = _pull_submatrix(K[iE], b, b)
        Kbi = _pull_submatrix(K[iE], b, i)
        Kib = _pull_submatrix(K[iE], i, b)
        Kii = _pull_submatrix(K[iE], i, i)
        Kii_inv = np.linalg.inv(Kii)
        Kbb -= Kbi @ Kii_inv @ Kib
        _push_submatrix(K_out[iE], Kbb, b, b)
        # load vector
        fb = _pull_subvector(f[iE], b)
        fi = _pull_subvector(f[iE], i)
        fb -= Kbi @ Kii_inv @ fi
        _push_subvector(f_out[iE], fb, b)
    return K_out, f_out


@njit(nogil=True, parallel=True, cache=__cache)
def condensate_M_bulk(M: ndarray, fixity: ndarray) -> ndarray:
    """
    Returns the condensed coefficient matrices representing constraints
    on the internal forces of the elements (eg. hinges).

    Parameters
    ----------
    M : numpy.ndarray
        3d float array, the mass matrices of several elements of the same kind.

    fixity: numpy.ndarray
        2d float boolean of connectivity factors for each dof of several elements.

    Returns
    -------
    numpy.ndarray
        The condensed mass matrices with the same shape as `M`.
        
    """
    nE = M.shape[0]
    M_out = np.zeros_like(M)
    for iE in prange(nE):
        b = np.where(fixity[iE])[0]
        i = np.where(~fixity[iE])[0]
        Mbb = _pull_submatrix(M[iE], b, b)
        Mbi = _pull_submatrix(M[iE], b, i)
        Mib = _pull_submatrix(M[iE], i, b)
        Mii = _pull_submatrix(M[iE], i, i)
        Mbb -= Mbi @ np.linalg.inv(Mii) @ Mib
        _push_submatrix(M_out[iE], Mbb, b, b)
    return M_out


def assert_min_diagonals_bulk(K: ndarray, minval: float = 1e-12):
    """
    Guarantees, that the values in the diagonal are larger
    a prescribed minimum value.

    Parameters
    ----------
    A : numpy.ndarray
        The coefficient matrix of several elements as a 3d
        array, where elements run along the first axis.

    minval : float, Optional
        The minimum value. Default is 1e-12.

    Returns
    -------
    numpy.ndarray
        The modified coefficient matrix, with the same shape as 'K'. 

    """
    inds = np.arange(K.shape[-1])
    d = K[:, inds, inds]
    eid, vid = np.where(d < minval)
    K[eid, vid, vid] = minval
    return K


def box_fem_data_sparse(K_coo: coo_matrix, Kp_coo: coo_matrix, f: ndarray):
    """
    Notes:
    ------
    If the load vector 'f' is dense, it must contain values for all
    nodes, even the passive ones.
    
    """
    # data for boxing and unboxing
    loc_to_glob, glob_to_loc = index_mappers(K_coo, return_inverse=True)
    # boxing
    K = box_spmatrix(K_coo, glob_to_loc) + box_spmatrix(Kp_coo, glob_to_loc)
    f = box_rhs(matrixform(f), loc_to_glob)
    return K, f, loc_to_glob


def box_fem_data_bulk(Kp_coo: coo_matrix, gnum: ndarray, f: ndarray):
    """
    Notes:
    ------
    If the load vector 'f' is dense, it must contain values for all
    nodes, even the passive ones.
    
    """
    # data for boxing and unboxing
    N = f.shape[0]
    loc_to_glob, glob_to_loc = index_mappers(gnum, N=N, return_inverse=True)
    # boxing
    gnum = box_dof_numbering(gnum, glob_to_loc)
    Kp_coo = box_spmatrix(Kp_coo, glob_to_loc)
    f = box_rhs(matrixform(f), loc_to_glob)
    return Kp_coo, gnum, f, loc_to_glob


# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
import numpy as np
from scipy.sparse import coo_matrix as coo, spmatrix, \
    isspmatrix as isspmatrix_np
from scipy.sparse.linalg import spsolve, spsolve_triangular, splu
from typing import Union
from time import time
#from concurrent.futures import ThreadPoolExecutor

from linkeddeepdict import LinkedDeepDict
from neumann.array import matrixform

from .imap import unbox_lhs
from .preproc import fem_coeff_matrix_coo, box_fem_data_sparse
from ..config import __haspardiso__

if __haspardiso__:
    import pypardiso as ppd
    from pypardiso import PyPardisoSolver
    #from pypardiso.scipy_aliases import pypardiso_core

arraylike = Union[ndarray, spmatrix]


def solve_standard_form(K: coo, f: np.ndarray, *args, use_umfpack:bool=True, summary:bool=False,
                        permc_spec:str='COLAMD', solver:str=None, mtype:int=11, assume_regular:bool=False,
                        **kwargs):
    """
    Solves the discrete equilibrium equations :math:`\mathbf{K} \mathbf{u} = \mathbf{f}`
    for :math:`\mathbf{u}`.
    
    Parameters
    ----------
    K : scipy.sparse.spmatrix
        The stiffness matrix in coo format.
        
    f : numpy.ndarray
        The vector of nodal loads.
        
    use_umfpack : bool, Optional
        Only if solver is 'scipy'. See the documentation :func:`scipy.sparse.linalg.spsolve` 
        for further meaning. Default is True.
        
    permc_spec : str, Optional
        Only if solver is 'scipy'. See the documentation :func:`scipy.sparse.linalg.spsolve` 
        for further meaning.
        
    solver : str, Optional
        The solver to use. Currently supported options are 'scipy' and 'pardiso'. If nothing is
        specified, we prefer 'pardiso' if it is around, otherwise the solver falls back to SciPy.
    
    mtype : int, Optional
        Matrix type indicator for the pypardiso solver. The Default value is 11, which denassumes
        a real and nonsymmetric coefficient matrix.
    
    assume_regular : bool, Optional
        If True, it is assumed that the input is according to the standard form, otherwise
        some preprocessing on the inputs is performed.
    
    Returns
    -------
    numpy.ndarray
        The solution as a numpy array. The returned array has the same shape as `f`.
        
    dict, Optional
        A summary including the most important characteristics of the solution process.
        Only if 'summary' is True.
        
    See also
    --------
    :func:`scipy.sparse.linalg.spsolve`
    :func:`scipy.sparse.linalg.splu`
        
    """
    solver = 'pardiso' if __haspardiso__ else 'scipy' if solver is None else solver
    if solver == 'pardiso' and not __haspardiso__:
        raise ImportError(
            "You need to install 'pypardiso' for solver type <{}>".format(solver))

    if not assume_regular:
        K.eliminate_zeros()
        K.sum_duplicates()
        f = matrixform(f)

    if solver == 'pardiso':
        if mtype == 11:
            t0 = time()
            u = ppd.spsolve(K, f)
            dt = time() - t0
        else:
            pds = PyPardisoSolver(mtype=mtype)
            t0 = time()
            u = pds.solve(K, f)
            dt = time() - t0
    elif solver == 'scipy':
        if len(args) > 0:
            if 'lower' in args:
                solver = 'scipy.sparse.linalg.spsolve_triangular (lower)'
                t0 = time()
                u = spsolve_triangular(K, f, lower=True)
                dt = time() - t0
            elif 'upper' in args:
                solver = 'scipy.sparse.linalg.spsolve_triangular (upper)'
                t0 = time()
                u = spsolve_triangular(K, f, lower=False)
                dt = time() - t0
            elif 'SLU' in args:
                solver = 'scipy.sparse.linalg.SuperLU'
                K = K.tocsc()
                t0 = time()
                LU = splu(K)
                u = LU.solve(f)
                dt = time() - t0
        else:
            if use_umfpack:
                use_umfpack = True if f.shape[1] == 1 else False
            solver = 'scipy.sparse.linalg.spsolve'
            t0 = time()
            u = spsolve(K, f, permc_spec=permc_spec, use_umfpack=use_umfpack)
            dt = time() - t0
    else:
        raise NotImplementedError(
            "Selected solver '{}' is not supported!".format(solver))

    if isspmatrix_np(u):
        u = u.todense()
    u = u.reshape(f.shape)

    # residual
    if summary:
        d = LinkedDeepDict(**{
            'time': dt,
            'N': u.shape[0],
            'solver': solver,
            'options': {}
            })
        if solver == 'scipy':
            d['options']['use_umfpack'] = use_umfpack
            d['options']['permc_spec'] = permc_spec
        elif solver == 'pardiso':
            d['options']['mtype'] = mtype
        return u, d
    else:
        return u


def linsolve_sparse(K_coo: coo, Kp_coo: coo,
                    f: np.ndarray, *args, use_umfpack=True,
                    sparsify=False, summary=True, **kwargs):
    """
    Notes:
    ------
    If the load vector 'f' is dense, it must contain values for all
    nodes, even the passive ones.
    
    """
    # boxing
    N = f.shape[0]
    K, f, loc_to_glob = box_fem_data_sparse(K_coo, Kp_coo, f)
    # solution
    u, d = solve_standard_form(K, f, *args, sparsify=sparsify,
                               use_umfpack=use_umfpack,
                               summary=True, **kwargs)
    # unboxing
    u = unbox_lhs(u, loc_to_glob, N)

    if summary:
        d['regular'] = loc_to_glob is not None
        return u, d
    return u


def linsolve_bulk(K_bulk: np.ndarray, Kp_coo: coo, f: np.ndarray,
                  gnum: np.ndarray, *args, **kwargs):
    K_coo = fem_coeff_matrix_coo(K_bulk, inds=gnum, N=f.shape[0])
    return linsolve_sparse(K_coo, Kp_coo, f, *args, **kwargs)


def linsolve(K: arraylike, *args, gnum=None, **kwargs):
    if isspmatrix_np(K):
        return linsolve_sparse(K, *args, **kwargs)
    elif isinstance(K, np.ndarray) and gnum is not None:
        return linsolve_bulk(K, *args, gnum, **kwargs)


"""def linsolve_bulk_pop(K_bulk: np.ndarray, Kp_coo: coo, f: np.ndarray,
                      gnum: np.ndarray, fcores: np.ndarray, *args,
                      sparsify=True, use_umfpack=True,
                      summary=False, parallel=True,
                      max_workers=4, permc_spec='COLAMD', **kwargs):
    f = matrixform(f)
    N = f.shape[0]
    loc_to_glob, glob_to_loc = \
        index_mappers(gnum, N=N, return_inverse=True)
    nPop, nE = fcores.shape
    gnum = box_dof_numbering(gnum, glob_to_loc)
    M = gnum.max() + 1
    dtype = K_bulk.dtype
    shape = (M, M)

    # rhs, lhs
    u = np.zeros((nPop,) + f.shape, dtype=dtype)
    f = box_rhs(f.astype(dtype), loc_to_glob)
    Kp_coo = box_spmatrix(Kp_coo, glob_to_loc)

    # stiffness utils
    rows, cols = list(map(lambda x: x.flatten(), irows_icols_bulk(gnum)))

    def K_bulk_w(i): return weighted_stiffness_bulk(
        K_bulk, fcores[i]).flatten()
    def K_coo(i): return coo((K_bulk_w(i), (rows, cols)), shape=shape,
                             dtype=dtype)

    def K(i): return K_coo(i) + Kp_coo

    def solve_and_unbox(i):
        ui = solve_standard_form(K(i), f, *args, sparsify=sparsify,
                                 use_umfpack=use_umfpack, permc_spec=permc_spec,
                                 **kwargs)
        u[i, :, :] = unbox_lhs(ui, loc_to_glob, N)

    # serial and parallel solutions
    if not parallel:
        t0 = time()
        list(map(solve_and_unbox, range(nPop)))
        dt = time() - t0
    else:
        max_workers = min(max_workers, nPop)
        t0 = time()
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            executor.map(solve_and_unbox, range(nPop))
        dt = time() - t0

    if summary:
        d = {
            'time [ms]': dt * 1000,
            'avg. time [ms]': dt * 1000 / nPop,
            'N': M,
            'use_umfpack': use_umfpack,
            'permc_spec': permc_spec,
            'sparsify': sparsify,
            'core': 'scipy.sparse.linalg.spsolve',
            'max workers': max_workers if parallel else 1,
            'population size': nPop
        }
        return u, d
    return u
"""



# -*- coding: utf-8 -*-
import numpy as np

import tetgen

from sectionproperties.analysis.section import Section

from neumann.linalg import linspace, Vector

from polymesh import PolyData
from polymesh.grid import grid
from polymesh.tri.triang import triangulate
from polymesh.space import StandardFrame, PointCloud
from polymesh.utils import centralize, center_of_points
from polymesh.tri.triutils import get_points_inside_triangles, \
    approx_data_to_points
from polymesh.topo import remap_topo
from polymesh.topo.tr import T6_to_T3

from sigmaepsilon.solid.fem import FemMesh
from sigmaepsilon.solid.fem.cells import TET4, H8

from ..fem.structure.structure import Structure


nodes_of_edges = [
    [0, 1], [1, 2], [2, 3], [3, 0],
    [4, 5], [5, 6], [6, 7], [7, 4],
    [0, 4], [1, 5], [2, 6], [3, 7]
]
corners_of_faces = {
    'front': [1, 2, 6, 5],
    'back': [0, 3, 7, 4],
    'left': [2, 3, 7, 6],
    'right': [0, 1, 5, 4],
    'bottom': [0, 1, 2, 3],
    'top': [4, 5, 6, 7],
}
edges_of_faces = {
    'front': [1, 5, 9, 10],
    'back': [3, 7, 8, 11],
    'right': [0, 9, 4, 8],
    'left': [2, 6, 10, 11],
    'bottom': [0, 1, 2, 3],
    'top': [4, 5, 6, 7],
}


def generate_frames(size, points_per_edge):
    points = []
    Lx, Ly, Lz = size
    # corners
    corner_coords = [
        [-Lx/2, -Ly/2, -Lz/2],
        [Lx/2, -Ly/2, -Lz/2],
        [Lx/2, Ly/2, -Lz/2],
        [-Lx/2, Ly/2, -Lz/2],
        [-Lx/2, -Ly/2, Lz/2],
        [Lx/2, -Ly/2, Lz/2],
        [Lx/2, Ly/2, Lz/2],
        [-Lx/2, Ly/2, Lz/2]
    ]
    corner_coords = np.array(corner_coords)
    points.append(corner_coords)

    # populate edges
    edge_coords = []
    N = points_per_edge + 2
    for nodes in nodes_of_edges:
        p0 = corner_coords[nodes[0]]
        p1 = corner_coords[nodes[1]]
        edge_coords.append(linspace(p0, p1, N)[1:-1])
    edge_coords = np.vstack(edge_coords)
    points.append(edge_coords)

    # center of face
    def cof(id):
        return center_of_points(corner_coords[corners_of_faces[id]])

    # face frames
    frames = {}
    Frame = StandardFrame(dim=3)
    frames['front'] = Frame.fork().move(cof('front'))
    rot90z = 'Body', [0, 0, np.pi/2], 'XYZ'
    frames['left'] = frames['front'].fork(
        *rot90z).move(cof('left') - cof('front'))
    frames['back'] = frames['left'].fork(
        *rot90z).move(cof('back') - cof('left'))
    frames['right'] = frames['back'].fork(
        *rot90z).move(cof('right') - cof('back'))
    rot_front_top = 'Body', [0, -np.pi/2, 0], 'XYZ'
    frames['top'] = frames['front'].fork(
        *rot_front_top).move(cof('top') - cof('front'))
    rot180y = 'Body', [0, np.pi, 0], 'XYZ'
    frames['bottom'] = frames['top'].fork(
        *rot180y).move(cof('bottom') - cof('top'))

    return points, frames


def joint_cube_voxelize(size, shape, *args, sections_dict=None,
                        material=None, **kwargs) -> Structure:
    points_per_edge = shape + 1
    mesh_size = min(size) / (points_per_edge-1)
    E = material['E']
    nu = material['nu']

    Frame = StandardFrame(dim=3)
    coords, topo = grid(size=size, shape=(shape, shape, shape),
                        eshape='H8', centralize=True)
    fixity = np.zeros_like(coords).astype(bool)
    loads = np.zeros_like(coords).astype(float)
    coords = PointCloud(coords, frame=Frame)
    _, frames = generate_frames(size, points_per_edge)

    for face in sections_dict:
        f_frame = frames[face]
        coords_ = coords.show(f_frame)
        f_inds = np.where(np.abs(coords_[:, 0]) < 1e-8)[0]
        f_coords = coords_[f_inds, 1:]
        f_section = sections_dict[face]['mesh']
        f_section.create_mesh(mesh_sizes=[mesh_size])
        f_coords_s = centralize(np.array(f_section.mesh['vertices']))
        f_topo_s = np.array(f_section.mesh['triangles'])

        # boundary conditions
        if 'dynams' in sections_dict[face]:
            dyn = sections_dict[face]['dynams']
            _section = Section(f_section)
            _section.calculate_meshetric_properties()
            _section.calculate_warping_properties()
            stress_post = _section.calculate_stress(
                N=dyn['N'], Vy=dyn['Vy'], Vx=dyn['Vx'],
                Mzz=dyn['T'], Mxx=dyn['Mx'], Myy=dyn['My']
            )
            stresses = stress_post.get_stress()[0]
            f_data_s = np.zeros((len(f_coords_s), 3))
            f_data_s[:, 0] = stresses['sig_zz']
            f_data_s[:, 1] = stresses['sig_zx']
            f_data_s[:, 2] = stresses['sig_zy']
            f_data_s = Vector(f_data_s, frame=f_frame).show(Frame)
            f_coords_s, f_topo_s = T6_to_T3(f_coords_s, f_topo_s)
            f_data = approx_data_to_points(
                f_coords_s, f_topo_s, f_data_s, f_coords)
            loads[f_inds, :] = f_data
        elif 'support' in sections_dict[face]:
            cond = get_points_inside_triangles(
                f_coords_s, f_topo_s[:, :3], f_coords)
            f_inds = np.where(cond)[0]
            fixity[f_inds, :] = True
        else:
            raise NotImplementedError

    # Hooke model
    Hooke = np.array([
        [1, nu, nu, 0, 0, 0],
        [nu, 1, nu, 0, 0, 0],
        [nu, nu, 1, 0, 0, 0],
        [0., 0, 0, (1-nu)/2, 0, 0],
        [0., 0, 0, 0, (1-nu)/2, 0],
        [0., 0, 0, 0, 0, (1-nu)/2]]) * (E / (1-nu**2))

    # return finite element model
    mesh = FemMesh(coords=coords.array, topo=topo, celltype=H8,
                   model=Hooke, fixity=fixity, loads=loads)
    structure = Structure(mesh=mesh)

    return structure


def joint_cube(size, shape, *args, sections_dict=None,
               voxelize=False, varvolume=None, material=None,
               mesh_ratio=1.0, **kwargs) -> Structure:
    """
    Returns a tetrahedral mesh of a cube, with sections optionally
    mounted on the faces.

    size : tuple
        Edge lengths of a cube in x, y and z directions.

    shape : int
        Number of line segments along an edge of the cube.

    sections_dict : dict
        A dictionary containing all information required to produce 
        meshes for the sections.

    material : dict, Optional
        A dictionary of material parameters for the cube. Only required
        if the input suggests a FEM calculation. Default is False.

    voxelize : bool, Optional
        Returns a voxelized mesh if set to `True`. Default is False.

    varvolume : float, Optional
        Controls the max volume of a cell in the resulting mesh
        of tetrahedra.

    Notes
    -----
    Only isotropic material definitions are supported here.

    Example
    -------
    >>> from sectionproperties.pre.library.steel_sections import circular_hollow_section as CHS
    >>> from sectionproperties.pre.library.steel_sections import rectangular_hollow_section as RHS
    >>> from sectionproperties.pre.library.steel_sections import i_section as ISection
    >>> from sectionproperties.analysis.section import Section
    >>> from sectionproperties.pre.pre import Material
    >>> import numpy as np

    A material must only be defined for a section if there are dynams 
    (forces or moments) defined on it, which requires the calculation of stresses.

    >>> E = 200e3
    >>> nu = 0.3
    >>> steel = Material(
    >>>     name='Steel', elastic_modulus=E, poissons_ratio=nu, density=7.85e-6,
    >>>     yield_strength=250, color='grey'
    >>> )

    >>> sections_dict = {
    >>>     'left': {
    >>>         'mesh': CHS(d=100, t=10, n=64),
    >>>         'support': True},
    >>>     'right': {
    >>>         'mesh': RHS(d=100, b=100, t=10, r_out=0, n_r=0),
    >>>         'support': True},
    >>>     'front': {
    >>>         'mesh': ISection(d=170, b=110, t_f=7.8, t_w=5.8, r=8.9, n_r=16, material=steel),
    >>>         'dynams': {'N': 1e3, 'Vx': 0., 'Vy': 3e3,
    >>>                    'T': 0.,  'Mx': 5e6, 'My': 0.}},
    >>> }

    >>> Lx, Ly, Lz = 150, 150, 150
    >>> voxelize = False

    >>> cube = joint_cube((Lx, Ly, Lz), 20, sections_dict=sections_dict,
    >>>                   material={'E': E, 'nu': nu}, voxelize=voxelize,
    >>>                   varvolume=60)

    """
    if voxelize:
        return joint_cube_voxelize(size, shape, *args,
                                   sections_dict=sections_dict,
                                   material=material, **kwargs)
    Lx, Ly, Lz = size
    area = Lx * Ly
    points_per_edge = shape + 1
    # size of line segments along an edge of the cube
    mesh_size = min(size) / (points_per_edge-1)
    nEdge = points_per_edge - 1  # number of line segments along an edge of the cube

    # base points
    points, frames = generate_frames(size, points_per_edge)
    points = [np.vstack(points), ]
    loads = [np.zeros((points[0].shape[0], 3)).astype(float), ]
    fixity = [np.zeros((points[0].shape[0], 3)).astype(bool), ]
    corner_coords = points[0][:8]
    edge_coords = points[0][8:]
    nTotalPoints = len(points[0])

    # background grid
    Frame = StandardFrame(dim=3)
    N = points_per_edge + 2
    coords_grid, topo_grid = grid(size=(Lx*0.99, Ly*0.99), shape=(N, N),
                                  eshape='Q4', centralize=True)
    Grid = PolyData(coords=coords_grid, topo=topo_grid, frame=Frame)
    grid_centers = Grid.centers()[:, :2]
    fixity_grid = np.zeros((grid_centers.shape[0], 3)).astype(bool)
    sig_grid = np.zeros((grid_centers.shape[0], 3)).astype(float)

    # loop over each face, add new points and triangulations
    for face in frames:
        f_frame = frames[face]
        # collect points on corners and edges and their global indices
        # these indices are used later to remap the default topology
        # resulting from per-face triangulations
        f_coords_base = []
        f_inds_base = []
        _corner_inds = []
        for corner in corners_of_faces[face]:
            f_coords_base.append(corner_coords[corner])
            _corner_inds.append(corner)
        f_inds_base.append(np.array(_corner_inds, dtype=int))
        for edge in edges_of_faces[face]:
            inds = np.arange(points_per_edge) + edge * points_per_edge
            f_coords_base.append(edge_coords[inds])
            f_inds_base.append(inds + 8)
        f_coords_base = np.vstack(f_coords_base)
        sig_coords_base = np.zeros((f_coords_base.shape[0], 3)).astype(float)
        fixity_coords_base = np.zeros((f_coords_base.shape[0], 3)).astype(bool)

        # transform the coords (base coordinates) so far to face frame
        f_coords_base = PointCloud(
            f_coords_base, frame=Frame).show(f_frame)

        # global indices and number of corner and edge points
        f_inds_base = np.concatenate(f_inds_base)
        nBasePoints = len(f_inds_base)

        # build face
        if face in sections_dict:
            if not 'mesh' in sections_dict[face]:
                continue
            # 1) create the mesh of the section
            # 2) rule out points of the base grid that the section covers
            # 3) add corner and edge nodes and do a triangulation
            if 'mesh_size' in sections_dict[face]:
                section_mesh_size = sections_dict[face]['mesh_size']
            else:
                section_mesh_size = mesh_size**2 * np.sqrt(3) / (4 * mesh_ratio)
                #section_mesh_size = area / (2 * mesh_ratio * nEdge**2)
            f_section = sections_dict[face]['mesh']
            f_section.create_mesh(mesh_sizes=[section_mesh_size])
            f_coords = centralize(np.array(f_section.mesh['vertices']))

            # boundary conditions
            sig_coords = np.zeros((f_coords.shape[0], 3)).astype(float)
            fixity_coords = np.zeros((f_coords.shape[0], 3)).astype(bool)
            if 'dynams' in sections_dict[face]:
                dyn = sections_dict[face]['dynams']
                _section = Section(f_section)
                _section.calculate_geometric_properties()
                _section.calculate_warping_properties()
                stress_post = _section.calculate_stress(
                    N=dyn['N'], Vy=dyn['Vy'], Vx=dyn['Vx'],
                    Mzz=dyn['T'], Mxx=dyn['Mx'], Myy=dyn['My']
                )
                stresses = stress_post.get_stress()
                sig_coords[:, 0] = stresses[0]['sig_zz']
                sig_coords[:, 1] = stresses[0]['sig_zx']
                sig_coords[:, 2] = stresses[0]['sig_zy']
                sig_coords = Vector(sig_coords, frame=f_frame).show(Frame)
            elif 'support' in sections_dict[face]:
                fixity_coords[:, :] = True
            else:
                raise NotImplementedError

            f_topo = np.array(f_section.mesh['triangles'].tolist())[:, :3]
            f_inds = get_points_inside_triangles(
                f_coords, f_topo, grid_centers)
            f_coords = np.vstack(
                [f_coords_base[:, 1:], f_coords, grid_centers[~f_inds]])
            f_coords, f_topo, _ = triangulate(points=f_coords)
            f_sig = np.vstack([sig_coords_base, sig_coords, sig_grid[~f_inds]])
            f_fixity = np.vstack(
                [fixity_coords_base, fixity_coords, fixity_grid[~f_inds]])
        else:
            f_coords = np.vstack([f_coords_base[:, 1:], grid_centers])
            f_coords, f_topo, _ = triangulate(points=f_coords)
            f_sig = np.zeros((f_coords.shape[0], 3)).astype(float)
            f_fixity = np.zeros((f_coords.shape[0], 3)).astype(bool)

        # faces share some points, hence they must be consistent
        # in node numbering --> remap topology to match indices
        # of corner and edge nodes
        f_inds = np.zeros(len(f_coords), dtype=int)
        nNewPoints = len(f_coords) - nBasePoints
        f_inds[:nBasePoints] = f_inds_base
        f_inds[nBasePoints:] = np.arange(nNewPoints) + nTotalPoints
        nTotalPoints += nNewPoints
        f_topo = remap_topo(f_topo, f_inds)

        # transform to global and append new data to total collection
        f_coords_new = np.zeros((nNewPoints, 3))
        f_coords_new[:, 1:] = f_coords[nBasePoints:]
        f_coords = PointCloud(f_coords_new, frame=f_frame).show(Frame)
        points.append(f_coords)
        fixity.append(f_fixity[nBasePoints:])
        loads.append(f_sig[nBasePoints:])

        # add topology to face to total collection
        if face not in sections_dict:
            sections_dict[face] = {}
        sections_dict[face]['topo'] = f_topo

    # build the final cube shell
    cubepoints = np.vstack(points)
    nCubePoints = len(cubepoints)
    cube = PolyData(coords=cubepoints, frame=Frame)
    for face in frames:
        if not 'topo' in sections_dict[face]:
            continue
        cube[face] = PolyData(topo=sections_dict[face]['topo'])

    # tetrahedralize
    switches = "pa{}".format(varvolume)
    tet = tetgen.TetGen(cube.coords(), cube.topology())
    tet.tetrahedralize(order=1, mindihedral=10, minratio=1.1,
                       quality=True, switches=switches, nobisect=True)
    tetgrid = tet.grid
    coords = np.array(tetgrid.points).astype(float)
    topo = tetgrid.cells_dict[10].astype(int)

    # collect loads and supports
    loads_ = np.zeros((coords.shape[0], 3)).astype(float)
    loads_[:nCubePoints] = np.vstack(loads)
    fixity_ = np.zeros((coords.shape[0], 3)).astype(bool)
    fixity_[:nCubePoints] = np.vstack(fixity)

    # Hooke model
    _E = material['E']
    nu = material['nu']
    Hooke = np.array([
        [1, nu, nu, 0, 0, 0],
        [nu, 1, nu, 0, 0, 0],
        [nu, nu, 1, 0, 0, 0],
        [0., 0, 0, (1-nu)/2, 0, 0],
        [0., 0, 0, 0, (1-nu)/2, 0],
        [0., 0, 0, 0, 0, (1-nu)/2]]) * (_E / (1-nu**2))

    # return finite element model
    mesh = FemMesh(coords=coords, topo=topo, celltype=TET4,
                   model=Hooke, fixity=fixity_, loads=loads_)
    structure = Structure(mesh=mesh)

    return structure


if __name__ == '__main__':
    from sectionproperties.pre.library.steel_sections import circular_hollow_section as CHS
    from sectionproperties.pre.library.steel_sections import rectangular_hollow_section as RHS
    from sectionproperties.pre.library.steel_sections import i_section as ISection
    from sectionproperties.analysis.section import Section
    from sectionproperties.pre.pre import Material
    import numpy as np
    import pyvista as pv

    E = 200e3
    nu = 0.3
    steel = Material(
        name='Steel', elastic_modulus=E, poissons_ratio=nu, density=7.85e-6,
        yield_strength=250, color='grey'
    )

    sections_dict = {
        'left': {
            'mesh': CHS(d=100, t=10, n=64),
            'support': True
        },
        'right': {
            'mesh': RHS(d=100, b=100, t=10, r_out=0, n_r=0),
            'support': True
        },
        'front': {
            'mesh': ISection(d=170, b=110, t_f=7.8, t_w=5.8,
                             r=8.9, n_r=16, material=steel),
            'dynams': {'N': 1e3, 'Vx': 0., 'Vy': 3e3,
                       'T': 0.,  'Mx': 5e6, 'My': 0.}
        },
    }

    sections_dict = {
        'left': {
            'mesh': CHS(d=100, t=10, n=64),
            'dynams': {'N': 0, 'Vx': 0., 'Vy': 0,
                       'T': 1e9,  'Mx': 0, 'My': 0.}
        },
        'right': {
            'mesh': RHS(d=100, b=100, t=10, r_out=0, n_r=0),
            'support': True},
    }

    Lx, Ly, Lz = 150, 150, 150
    voxelize = False

    cube = joint_cube((Lx, Ly, Lz), 40, sections_dict=sections_dict,
                      material={'E': E, 'nu': nu}, voxelize=voxelize,
                      varvolume=60)

    cube.plot()
    coords = cube.coords()
    topo = cube.topology()
    cube.linsolve()

    data = cube.stresses_at_centers('HMH')
    centers = cube.centers()
    dofsol = cube.dofsol(flatten=False)

    mask = np.where(centers[:, 0] > 0)[0]

    p = pv.Plotter(notebook=False)
    celltype = H8 if voxelize else TET4
    cube2 = PolyData(coords=coords + dofsol,
                     topo=topo[mask], celltype=celltype)
    p.add_mesh(cube2.to_pv(), scalars=data[mask], show_edges=True)
    p.add_mesh(cube.mesh.to_pv(), style='wireframe', color='black')
    p.show()
